"""Forms for configuring repositories."""

from __future__ import annotations

import logging
import sys
from itertools import chain
from typing import Optional, TYPE_CHECKING, Type, Union

from django import forms
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.forms import Select, model_to_dict
from django.utils.datastructures import MultiValueDict
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext, gettext_lazy as _
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.filesystem import is_exe_in_path

from reviewboard.admin.form_widgets import (RelatedGroupWidget,
                                            RelatedUserWidget)
from reviewboard.admin.import_utils import has_module
from reviewboard.admin.validation import validate_bug_tracker
from reviewboard.certs.errors import CertificateVerificationError
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceError,
                                            MissingHostingServiceError,
                                            SSHKeyAssociationError,
                                            TwoFactorAuthCodeRequiredError)
from reviewboard.hostingsvcs.base import hosting_service_registry
from reviewboard.hostingsvcs.fake import FAKE_HOSTING_SERVICES
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.reviews.models import Group
from reviewboard.scmtools import scmtools_registry
from reviewboard.scmtools.certs import Certificate as LegacyCertificate
from reviewboard.scmtools.errors import (
    AuthenticationError,
    RepositoryNotFoundError,
    SCMError,
    UnverifiedCertificateError as LegacyUnverifiedCertificateError)
from reviewboard.scmtools.fake import FAKE_SCMTOOLS
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.site.mixins import LocalSiteAwareModelFormMixin
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.ssh.client import SSHClient
from reviewboard.ssh.errors import (BadHostKeyError,
                                    SSHError,
                                    UnknownHostKeyError)

if TYPE_CHECKING:
    from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService


logger = logging.getLogger(__name__)


class HostingAccountWidget(Select):
    """A widget for selecting and modifying an assigned hosting account.

    This presents a list of available hosting service accounts as a drop-down,
    and provides a link for editing the credentials of the selected account.
    """

    def render(self, *args, **kwargs):
        """Render the widget.

        Args:
            *args (tuple):
                Arguments for the render.

            **kwargs (dict):
                Keyword arguments for the render.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for the widget.
        """
        html = super(HostingAccountWidget, self).render(*args, **kwargs)

        return mark_safe(html + (
            '<a href="#" id="repo-edit-hosting-credentials">'
            '<span class="rb-icon rb-icon-edit"></span> '
            '<span id="repo-edit-hosting-credentials-label">%s</span></a>'
            % _('Edit credentials')
        ))


class BaseRepositorySubForm(forms.Form):
    """A sub-form used in the main repository configuration form.

    This provides some standard functionality for collecting information
    needed to configure a specific type of repository (one backed by a
    particular :py:class:`~reviewboard.scmtools.core.SCMTool` or
    :py:class:`~reviewboard.hostingsvcs.base.hosting_service.
    BaseHostingService`). It takes care of basic form customization and
    loading, and must be subclassed for other operations.

    Third-parties will never need to subclass this directly. Instead, subclass
    one of:

    * :py:class:`~reviewboard.hostingsvcs.forms.HostingServiceForm`
    * :py:class:`~reviewboard.hostingsvcs.base.forms.
      BaseHostingServiceAuthForm`

    Forms can provide a :py:class:`Meta` class that define
    :py:attr:`Meta.help_texts` and :py:attr:`Meta.labels` attributes. Each is
    a dictionary mapping field names to new content for those fields. See the
    classes above for examples.

    Version Added:
        3.0.16

    Attributes:
        local_site (reviewboard.site.models.LocalSite):
            The :term:`Local Site` that any queries or state should be bound
            to.

        repository (reviewboard.scmtools.models.Repository):
            The repository being configured. This is allowed to be ``None``,
            mainly for testing purposes, but will always have a value when
            constructed by :py:class:`RepositoryForm`.
    """

    # Turn off client-side validation, performing validation only server-side.
    use_required_attribute = False

    def __init__(self, *args, **kwargs):
        """Initialize the form.

        Subclasses should use this to alter the fields shown in the form, if
        needed.

        Args:
            *args (tuple):
                Additional positional arguments for the parent form.

            **kwargs (dict):
                Additional keyword arguments for the parent form.

        Keyword Args:
            repository (reviewboard.scmtools.models.Repository, optional):
                The repository that's being created or updated. This is
                allowed to be ``None``, mainly for testing purposes, but will
                always have a value when constructed by
                :py:class:`RepositoryForm`.

            local_site (reviewboard.site.models.LocalSite, optional):
                The :term:`Local Site` that any queries or state should be
                bound to.
        """
        # Pull these out of kwargs so that we can preserve *args calls
        # without problems.
        self.repository = kwargs.pop('repository', None)
        self.local_site = kwargs.pop('local_site', None)

        super(BaseRepositorySubForm, self).__init__(*args, **kwargs)

        # Allow the subclass to override help text and field labels.
        meta = getattr(self, 'Meta', None)

        if meta is not None:
            help_texts = getattr(meta, 'help_texts', {})
            labels = getattr(meta, 'labels', {})

            for field_name, help_text in help_texts.items():
                self.fields[field_name].help_text = help_text

            for field_name, label in labels.items():
                self.fields[field_name].label = label

    def get_initial_data(self):
        """Return initial data for the form.

        By default, this doesn't return any initial data. Subclasses can
        override this to return something suitable for the form.

        Generally, sensitive information, like passwords, should not be
        provided.

        Returns:
            dict:
            Initial data for the form.
        """
        return {}

    def load(self):
        """Load information for the form.

        By default, this will populate initial values returned in
        :py:meth:`get_initial_data`. Subclasses can override this to set
        other fields or state as needed.
        """
        for key, value in self.get_initial_data().items():
            self.fields[key].initial = value

    def save(self):
        """Save information from the form.

        Subclasses must override this.
        """
        raise NotImplementedError

    def get_field_data_from(self, obj, field_names=None, model_fields=None,
                            norm_key_func=None):
        """Return data from an object for use in the form's fields.

        This is a utility method that helps load in field data based on the
        attributes on an object and the object's ``extra_data`` field. It's
        most commonly going to be used for a subclass's :py:meth:`load` or
        :py:meth:`get_initial_data`.

        Args:
            obj (django.db.models.Model):
                The model object to load data from. This is expected to have
                an ``extra_data`` field.

            field_names (list of unicode, optional):
                A specific list of field names to load from the object. If
                not provided, this defaults to the form's list of field names.
                These do not all have to be present in the object.

            model_fields (set of unicode, optional):
                Names of fields that should be loaded directly from attributes
                on ``obj``, instead of the object's ``extra_data``.

            norm_key_func (callable, optional):
                A function that normalizes a key before looking up in the
                object's ``extra_data``. If not provided, this defaults to
                :py:meth:`~django.forms.forms.BaseForm.add_prefix`.

        Returns:
            dict:
            The loaded field data.
        """
        data = {}
        model_fields = set(model_fields or [])

        if field_names is None:
            field_names = self.fields.keys()

        if norm_key_func is None:
            norm_key_func = self.add_prefix

        for key in field_names:
            if key in model_fields:
                data[key] = getattr(obj, key)
            else:
                norm_key = norm_key_func(key)

                if norm_key in obj.extra_data:
                    data[key] = obj.extra_data[norm_key]

        return data


class BaseRepositoryAuthSubForm(BaseRepositorySubForm):
    """Base class for any repository authentication forms.

    Third-parties will never need to subclass this directly. Instead, subclass
    one of:

    * :py:class:`~reviewboard.hostingsvcs.base.forms.
      BaseHostingServiceAuthForm`
    * :py:class:`~reviewboard.scmtools.forms.BaseSCMToolAuthForm`
    """


class BaseRepositoryInfoSubForm(BaseRepositorySubForm):
    """Base class for any repository information forms.

    Third-parties will never need to subclass this directly. Instead, subclass
    one of:

    * :py:class:`~reviewboard.hostingsvcs.forms.HostingServiceForm`
    * :py:class:`~reviewboard.scmtools.forms.BaseSCMToolRepositoryForm`
    """


class SCMToolSubFormMixin(object):
    """Mixin class for SCMTool-specific subforms.

    This should only be used internally. SCMTools will want to subclass
    :py:class:`BaseSCMToolAuthForm`, :py:class:`BaseSCMToolRepositoryForm`,
    or one of their descendents.

    Version Added:
        3.0.16

    Attributes:
        scmtool_cls (type):
            The :py:class:`~reviewboard.scmtools.core.SCMTool` subclass used
            for this form.
    """

    #: A set of fields to save directly to the Repository model.
    #:
    #: This should only be set by internal classes.
    _MODEL_FIELDS = set()

    #: A set of fields to save in extra_data without an SCMTool ID prefix.
    #:
    #: This exists for backwards-compatibility with older configuration,
    #: and should only be set by internal classes.
    _PREFIXLESS_KEYS = set()

    def __init__(self, **kwargs):
        """Initialize the form.

        Subclasses should use this to alter the fields shown in the form, if
        needed, but not to set initial form field values from the repository,
        as those will be overridden.

        Args:
            **kwargs (dict):
                Additional keyword arguments for the parent form.

        Keyword Args:
            scmtool_cls (type):
                The subclass of :py:class:`~reviewboard.scmtools.core.SCMTool`
                that this form represents.
        """
        scmtool_cls = kwargs.pop('scmtool_cls')
        self.scmtool_cls = scmtool_cls

        super(SCMToolSubFormMixin, self).__init__(**kwargs)

        for name, help_text in scmtool_cls.field_help_text.items():
            if name in self.fields:
                self.fields[name].help_text = help_text

    def get_initial_data(self):
        """Return initial data for the form.

        This will load information from the repository's attributes and
        :py:attr:`~reviewboard.scmtools.models.Repository.extra_data` into the
        form's fields.

        Returns:
            dict:
            Initial data for the form.
        """
        def _norm_key(key):
            if key in self._PREFIXLESS_KEYS:
                return key

            return self.add_prefix(key)

        return self.get_field_data_from(self.repository,
                                        model_fields=self._MODEL_FIELDS,
                                        norm_key_func=_norm_key)

    def save(self):
        """Save information to the repository.

        This will store the content of the fields in the repository.

        Subclasses will generally not need to override this.
        """
        repository = self.repository
        assert repository is not None

        for key, value in self.cleaned_data.items():
            if key in self._MODEL_FIELDS:
                setattr(repository, key, value)
            elif key in self._PREFIXLESS_KEYS:
                repository.extra_data[key] = value
            else:
                repository.extra_data[self.add_prefix(key)] = value

    def __repr__(self):
        """Return a string representation of the form.

        Args:
            unicode:
            The string representation.
        """
        return '<%s (scmtool=%s)>' % (type(self).__name__,
                                      self.scmtool_cls.scmtool_id)


class BaseSCMToolAuthForm(SCMToolSubFormMixin, BaseRepositoryAuthSubForm):
    """Base class for SCMTool authentication forms.

    This is a blank form that can be subclassed and populated with fields for
    requesting authentication credentials for plain repositories.

    Any cleaned data fields named ``username`` or ``password`` will be set
    directly on the equivalent
    :py:class:`~reviewboard.scmtools.models.Repository` model fields.  Any
    other fields will be stored in :py:attr:`Repository.extra_data
    <reviewboard.scmtools.models.Repository.extra_data>`, using a key in the
    form of :samp:`<scmtoolid>_<fieldname>`.

    If an SCMTool uses a standard username/password, they're most likely
    going to want to use :py:class:`StandardSCMToolAuthForm` directly or as
    a parent class.

    Version Added:
        3.0.16
    """

    _MODEL_FIELDS = {'username', 'password'}


class BaseSCMToolRepositoryForm(SCMToolSubFormMixin,
                                BaseRepositoryInfoSubForm):
    """Base class for SCMTool repository forms.

    This is a blank form that can be subclassed and populated with fields for
    requesting information for plain repositories.

    Subclasses are required to provide a :guilabel:`Path` field, or to at least
    provide a suitable value in the cleaned data based on other fields.

    Any cleaned data fields named ``path``, ``mirror_path``, or
    ``raw_file_url`` will be set directly on the equivalent
    :py:class:`~reviewboard.scmtools.models.Repository` model fields. Any
    other fields will be stored in :py:attr:`Repository.extra_data
    <reviewboard.scmtools.models.Repository.extra_data>`, using a key in the
    form of :samp:`<scmtoolid>_<fieldname>`. The exception is the field
    ``use_ticket_auth``, which will be stored without an SCMTool ID prefix for
    legacy reasons.

    If an SCMTool wants to provide standard path/mirror path fields, they're
    most likely going to want to use :py:class:`StandardSCMToolRepositoryForm`
    directly or as a parent class.

    Version Added:
        3.0.16
    """

    _MODEL_FIELDS = {'path', 'mirror_path', 'raw_file_url'}
    _PREFIXLESS_KEYS = {'use_ticket_auth'}


class StandardSCMToolAuthForm(BaseSCMToolAuthForm):
    """A standard SCMTool authentication form.

    This provides standard :guilabel:`Username` and :guilabel:`Password`
    fields. These are optional by default. Subclasses can override them to make
    the fields required, remove them, or add additional authentication-related
    fields.

    See the documentation on the :py:class:`parent class <BaseSCMToolAuthForm>`
    to see how field data is stored.

    Version Added:
        3.0.16
    """

    username = forms.CharField(
        max_length=Repository._meta.get_field('username').max_length,
        required=False,
        widget=forms.TextInput(attrs={
            'autocomplete': 'off',
            'size': '30',
        }))

    password = forms.CharField(
        label=_('Password'),
        required=False,
        widget=forms.PasswordInput(
            render_value=True,
            attrs={
                'autocomplete': 'off',
                'size': '30',
            }))

    def clean_username(self):
        """Clean the username field.

        This will strip all whitespace from the field before returning it.

        Returns:
            unicode:
            The value provided in the field, with whitespace stripped.
        """
        return self.cleaned_data['username'].strip()

    def clean_password(self):
        """Clean the password field.

        This will strip all whitespace from the field before returning it.

        Returns:
            unicode:
            The value provided in the field, with whitespace stripped.
        """
        return self.cleaned_data['password'].strip()


class StandardSCMToolRepositoryForm(BaseSCMToolRepositoryForm):
    """A standard SCMTool repository form.

    This provides standard :guilabel:`Path` and :guilabel:`Mirror Path` fields,
    as well as optional fields for :guilabel:`Raw File URL Mask` (if
    :py:class:`SCMTool.raw_file_url
    <reviewboard.scmtools.core.SCMTool.supports_raw_file_urls>` is set) and
    :guilabel:`Use ticket-based authentication` <if
    :py:class:`SCMTool.raw_file_url
    <reviewboard.scmtools.core.SCMTool.supports_ticket_auth>` is set). These
    two optional fields are provided for legacy purposes, but will be removed
    in the future, so subclasses should explicitly provide them if needed.

    Subclasses can override any of the form's fields, remove them, or add
    additional fields needed to identify repositories.

    If a :guilabel:`Path` field is not appropriate for the type of repository,
    then it's still up to the subclass to provide a suitable ``path`` value
    in the cleaned data that uniquely identifies the repository.

    See the documentation on the :py:class:`parent class
    <BaseSCMToolRepositoryForm>` to see how field data is stored.

    Version Added:
        3.0.16
    """

    path = forms.CharField(
        label=_('Path'),
        max_length=Repository._meta.get_field('path').max_length,
        widget=forms.TextInput(attrs={'size': 60}))

    mirror_path = forms.CharField(
        label=_('Mirror Path'),
        required=False,
        max_length=Repository._meta.get_field('mirror_path').max_length,
        widget=forms.TextInput(attrs={'size': 60}))

    raw_file_url = forms.CharField(
        label=_('Raw File URL Mask'),
        max_length=Repository._meta.get_field('raw_file_url').max_length,
        required=False,
        widget=forms.TextInput(attrs={'size': 60}),
        help_text=_("A URL mask used to check out a particular revision of a "
                    "file using HTTP. This is needed for repository types "
                    "that can't access remote files natively. "
                    "Use <tt>&lt;revision&gt;</tt> and "
                    "<tt>&lt;filename&gt;</tt> in the URL in place of the "
                    "revision and filename parts of the path."))

    use_ticket_auth = forms.BooleanField(
        label=_('Use ticket-based authentication'),
        initial=False,
        required=False)

    def __init__(self, **kwargs):
        """Initialize the form.

        This will set the appropriate fields on the form based on the
        capabilities on the :py:class:`~reviewboard.scmtools.core.SCMTool`,
        as per the class's documentation.

        Args:
            **kwargs (dict):
                Additional keyword arguments for the parent form.
        """
        super(StandardSCMToolRepositoryForm, self).__init__(**kwargs)

        if not self.scmtool_cls.supports_raw_file_urls:
            del self.fields['raw_file_url']

        if not self.scmtool_cls.supports_ticket_auth:
            del self.fields['use_ticket_auth']

    def clean_path(self):
        """Clean the Path field.

        This will strip all whitespace from the field before returning it.

        Returns:
            unicode:
            The value provided in the field, with whitespace stripped.
        """
        path = self.cleaned_data['path'].strip()

        if not path:
            raise ValidationError(gettext('Repository path cannot be empty'))

        return path

    def clean_mirror_path(self):
        """Clean the Mirror Path field.

        This will strip all whitespace from the field before returning it.

        Returns:
            unicode:
            The value provided in the field, with whitespace stripped.
        """
        return self.cleaned_data['mirror_path'].strip()

    def clean_raw_file_url(self):
        """Clean the Raw File URL Mask field.

        This will strip all whitespace from the field before returning it.

        Returns:
            unicode:
            The value provided in the field, with whitespace stripped.
        """
        return self.cleaned_data['raw_file_url'].strip()


class RepositoryForm(LocalSiteAwareModelFormMixin, forms.ModelForm):
    """A form for creating and updating repositories.

    This form provides an interface for creating and updating repositories,
    handling the association with hosting services, linking accounts,
    dealing with SSH keys and SSL certificates, and more.

    Configuration details are collected primarily through subforms provided
    by SCMTools and Hosting Services.
    """

    REPOSITORY_HOSTING_FIELDSET = _('Repository Hosting')
    REPOSITORY_INFO_FIELDSET = _('Repository Information')
    BUG_TRACKER_FIELDSET = _('Bug Tracker')
    SSH_KEY_FIELDSET = _('Review Board Server SSH Key')

    NO_HOSTING_SERVICE_ID = 'custom'
    NO_HOSTING_SERVICE_NAME = _('(None - Custom Repository)')

    NO_BUG_TRACKER_ID = 'none'
    NO_BUG_TRACKER_NAME = _('(None)')

    CUSTOM_BUG_TRACKER_ID = 'custom'
    CUSTOM_BUG_TRACKER_NAME = _('(Custom Bug Tracker)')

    IGNORED_SERVICE_IDS = ('none', 'custom')

    DEFAULT_PLAN_ID = 'default'
    DEFAULT_PLAN_NAME = _('Default')

    _SCMTOOL_PREFIXLESS_FIELDS = (BaseSCMToolAuthForm._MODEL_FIELDS |
                                  BaseSCMToolAuthForm._PREFIXLESS_KEYS |
                                  BaseSCMToolRepositoryForm._MODEL_FIELDS |
                                  BaseSCMToolRepositoryForm._PREFIXLESS_KEYS)

    # Turn off client-side validation, performing validation only server-side.
    use_required_attribute = False

    # Host trust state
    reedit_repository = forms.BooleanField(
        label=_("Re-edit repository"),
        required=False)

    trust_host = forms.BooleanField(
        label=_("I trust this host"),
        required=False)

    # Repository Hosting fields
    hosting_type = forms.ChoiceField(
        label=_("Hosting service"),
        required=True,
        initial=NO_HOSTING_SERVICE_ID)

    hosting_account = forms.ModelChoiceField(
        label=_('Account'),
        required=True,
        empty_label=_('<Link a new account>'),
        help_text=_("Link this repository to an account on the hosting "
                    "service. This username may be used as part of the "
                    "repository URL, depending on the hosting service and "
                    "plan."),
        queryset=(
            HostingServiceAccount.objects
            .accessible(local_site=LocalSite.ALL)
        ),
        widget=HostingAccountWidget())

    force_authorize = forms.BooleanField(
        label=_('Force reauthorization'),
        required=False,
        widget=forms.HiddenInput())

    # Repository Information fields
    tool = forms.ChoiceField(
        label=_("Repository type"),
        required=True)

    repository_plan = forms.ChoiceField(
        label=_('Repository plan'),
        required=True,
        help_text=_('The plan for your repository on this hosting service. '
                    'This must match what is set for your repository.'))

    # Auto SSH key association field
    associate_ssh_key = forms.BooleanField(
        label=_('Associate my SSH key with the hosting service'),
        required=False,
        help_text=_('Add the Review Board public SSH key to the list of '
                    'authorized SSH keys on the hosting service.'))

    NO_KEY_HELP_FMT = (_('This repository type supports SSH key association, '
                         'but the Review Board server does not have an SSH '
                         'key. <a href="%s">Add an SSH key.</a>'))

    # Bug Tracker fields
    bug_tracker_use_hosting = forms.BooleanField(
        label=_("Use hosting service's bug tracker"),
        initial=False,
        required=False)

    bug_tracker_type = forms.ChoiceField(
        label=_("Type"),
        required=True,
        initial=NO_BUG_TRACKER_ID)

    bug_tracker_hosting_url = forms.CharField(
        label=_('URL'),
        required=True,
        widget=forms.TextInput(attrs={'size': 30}))

    bug_tracker_plan = forms.ChoiceField(
        label=_('Bug tracker plan'),
        required=True)

    bug_tracker_hosting_account_username = forms.CharField(
        label=_('Account username'),
        required=True,
        widget=forms.TextInput(attrs={'size': 30, 'autocomplete': 'off'}))

    bug_tracker = forms.CharField(
        label=_("Bug tracker URL"),
        max_length=256,
        required=False,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=(
            _("The optional path to the bug tracker for this repository. The "
              "path should resemble: http://www.example.com/issues?id=%%s, "
              "where %%s will be the bug number.")
            % ()),  # We do this wacky formatting trick because otherwise
                    # xgettext gets upset that it sees a format string with
                    # positional arguments and will abort when trying to
                    # extract the message catalog.
        validators=[validate_bug_tracker])

    # Access control fields
    users = forms.ModelMultipleChoiceField(
        queryset=User.objects.filter(is_active=True),
        label=_('Users with access'),
        required=False,
        widget=RelatedUserWidget())

    review_groups = forms.ModelMultipleChoiceField(
        queryset=Group.objects.order_by('name'),
        label=_('Groups with access'),
        required=False,
        widget=RelatedGroupWidget(invite_only=True))

    ######################
    # Instance variables #
    ######################

    #: A legacy certificate accepted from legacy SCMTools.
    #:
    #: Type
    #:     reviewboard.scmtools.certs.Certificate
    cert: Optional[LegacyCertificate]

    #: A certificate verification exception generated during validation.
    #:
    #: Type:
    #:     reviewboard.certs.errors.CertificateVerificationError or
    #:     reviewboard.scmtools.errors.UnverifiedCertificateError
    certerror: Optional[Union[CertificateVerificationError,
                              LegacyUnverifiedCertificateError]]

    def __init__(self, data=None, *args, **kwargs):
        """Initialize the repository configuration form.

        This will set up the initial state for the form, locating any
        tools and hosting services that can be shown and setting up the
        configuration and authentication forms they provide.

        Args:
            data (dict, optional):
                The posted form data.

            *args (tuple):
                Positional arguments to pass to the parent class.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        from reviewboard.hostingsvcs.base.forms import \
            BaseHostingServiceAuthForm

        # Django's admin UI will pass RepositoryForm an immutable QueryDict
        # as the POST data. This normally makes sense for 99.9% of forms, but
        # this form is a bit special.
        #
        # Certain fields need to be updated when we calculate some data so
        # that we can set better defaults if the form doesn't fully save. For
        # instance, if we successfully link a new account but fail to validate
        # a repository, we want the default account to be the newly-linked
        # account, and this requires being able to modify self.data.
        if isinstance(data, MultiValueDict):
            data = data.dict()

        super(RepositoryForm, self).__init__(data, *args, **kwargs)

        self.hostkeyerror = None
        self.certerror = None
        self.userkeyerror = None
        self.form_validation_error = None

        self.hosting_account_linked = False
        self.hosting_bug_tracker_forms = {}
        self.hosting_auth_forms = {}
        self.hosting_repository_forms = {}
        self.hosting_service_info = {}

        self.scmtool_auth_forms = {}
        self.scmtool_repository_forms = {}
        self.scmtool_info = {}

        self.subforms_cleaned_data = None
        self.subforms_errors = None
        self.subforms_valid = False

        self.cert = None

        # Create some aliases for the current Local Site and name handled by
        # the LocalSiteAwareModelFormMixin. This may be a Local Site bound to
        # the form or one specified in form data (in which case it would have
        # already been checked for access rights).
        if self.cur_local_site is not None:
            self.local_site = self.cur_local_site
        else:
            self.local_site = None

        # Grab the entire list of HostingServiceAccounts that can be
        # used by this form. When the form is actually being used by the
        # user, the listed accounts will consist only of the ones available
        # for the selected hosting service.
        #
        # These will be fed into auth forms. We don't modify the queryset here,
        # since the LocalSiteAwareModelFormMixin will manage that for us.
        hosting_accounts = list(
            HostingServiceAccount.objects
            .accessible(local_site=self.cur_local_site)
        )

        # Standard forms don't support 'instance', so don't pass it through
        # to any created hosting service forms.
        if 'instance' in kwargs:
            kwargs.pop('instance')

        # Set some fields based on the instance. We only want to work with
        # fields here that aren't dependent on any loaded hosting service or
        # SCMTool forms or state.
        instance = self.instance
        cur_hosting_service_cls: (type[BaseHostingService] | None) = None

        if instance:
            cur_scmtool_cls = instance.scmtool_class

            try:
                cur_hosting_service_cls = type(instance.hosting_service)
            except MissingHostingServiceError:
                pass

            if cur_scmtool_cls is not None:
                self.fields['tool'].initial = cur_scmtool_cls.scmtool_id

            self._populate_hosting_service_fields()
            self._populate_bug_tracker_fields()

            # If the repository is public, but has access lists set (which
            # could happen prior to 3.0.16 if setting an access list and then
            # unchecking the Public Access checkbox), make sure we're not
            # reflecting those access lists here in the UI so there isn't any
            # confusion when toggling that checkbox. We want them to start
            # fresh.
            #
            # Saving will also clear out any access lists if set to public.
            if instance.public:
                # Note that because we loaded from an instance, the populated
                # values are in self.initial and not in field.initial.
                self.initial['users'] = []
                self.initial['review_groups'] = []

        # Load the list of repository forms and hosting services.
        hosting_service_choices = []
        bug_tracker_choices = []

        hosting_services = set()

        for hosting_service in hosting_service_registry:
            if (not hosting_service.visible and
                hosting_service is not cur_hosting_service_cls):
                # We don't want to show this service as an option.
                continue

            hosting_service_id = hosting_service.hosting_service_id
            hosting_services.add(hosting_service_id)

            auth_form_cls = \
                hosting_service.auth_form or BaseHostingServiceAuthForm

            if hosting_service.supports_repositories:
                hosting_service_choices.append(
                    (hosting_service_id, hosting_service.name)
                )

            if hosting_service.supports_bug_trackers:
                bug_tracker_choices.append(
                    (hosting_service_id, hosting_service.name)
                )

            self.hosting_bug_tracker_forms[hosting_service_id] = {}
            self.hosting_repository_forms[hosting_service_id] = {}
            self.hosting_service_info[hosting_service_id] = \
                self._get_hosting_service_info(
                    hosting_service=hosting_service,
                    hosting_accounts=hosting_accounts,
                    is_instance_service=(hosting_service is
                                         cur_hosting_service_cls))

            try:
                if hosting_service.plans:
                    for type_id, info in hosting_service.plans:
                        form = info.get('form', None)

                        if form:
                            self._load_hosting_service(
                                hosting_service_id=hosting_service_id,
                                hosting_service=hosting_service,
                                plan_type_id=type_id,
                                plan_type_label=info['name'],
                                form_class=form)
                elif hosting_service.form:
                    self._load_hosting_service(
                        hosting_service_id=hosting_service_id,
                        hosting_service=hosting_service,
                        plan_type_id=self.DEFAULT_PLAN_ID,
                        plan_type_label=self.DEFAULT_PLAN_NAME,
                        form_class=hosting_service.form)

                # Load the hosting service's custom authentication form.
                #
                # We start off constructing the form without any data. We
                # don't want to prematurely trigger any validation for forms
                # that we won't end up using (even if it matches the current
                # hosting time, as we still don't know if we need to link or
                # re-authorize an account). We'll replace this with a populated
                # form further below.
                #
                # Note that we do still need the form instantiated here, for
                # template rendering.
                self.hosting_auth_forms[hosting_service_id] = \
                    auth_form_cls(hosting_service_cls=hosting_service,
                                  repository=instance,
                                  local_site=self.local_site,
                                  prefix=hosting_service_id)
            except Exception as e:
                logger.exception('Error loading hosting service %s: %s',
                                 hosting_service_id, e)

        for hosting_service_id, cls in FAKE_HOSTING_SERVICES.items():
            if hosting_service_id not in hosting_services:
                service_info = self._get_hosting_service_info(cls)
                service_info['fake'] = True
                self.hosting_service_info[hosting_service_id] = \
                    service_info

                hosting_service_choices.append((hosting_service_id,
                                                cls.name))

        # Build the list of hosting service choices, sorted, with
        # "None" being first.
        hosting_service_choices.sort(key=lambda x: x[1])
        hosting_service_choices.insert(0, (self.NO_HOSTING_SERVICE_ID,
                                           self.NO_HOSTING_SERVICE_NAME))
        self.fields['hosting_type'].choices = hosting_service_choices

        # Now do the same for bug trackers, but have separate None and Custom
        # entries.
        bug_tracker_choices.sort(key=lambda x: x[1])
        bug_tracker_choices.insert(0, (self.NO_BUG_TRACKER_ID,
                                       self.NO_BUG_TRACKER_NAME))
        bug_tracker_choices.insert(1, (self.CUSTOM_BUG_TRACKER_ID,
                                       self.CUSTOM_BUG_TRACKER_NAME))
        self.fields['bug_tracker_type'].choices = bug_tracker_choices

        # Load the list of SCM tools.
        available_scmtools = set()
        scmtool_choices = []
        hosting_type_value = self['hosting_type'].value()
        tool_value = self['tool'].value()

        for scmtool in scmtools_registry:
            scmtool_id = scmtool.scmtool_id
            is_tool_active = (tool_value == scmtool_id and
                              hosting_type_value == self.NO_HOSTING_SERVICE_ID)

            try:
                self._load_scmtool(scmtool_cls=scmtool,
                                   is_active=is_tool_active)
            except Exception as e:
                logger.exception('Error loading SCMTool %s <%s>: %s',
                                 scmtool_id, scmtool.class_name, e)
                continue

            self.scmtool_info[scmtool_id] = self._get_scmtool_info(scmtool)
            scmtool_choices.append((scmtool_id, scmtool.name))
            available_scmtools.add(scmtool_id)

        # Create placeholders for any SCMTools we want to list that aren't
        # currently installed.
        for scmtool_id, name in FAKE_SCMTOOLS.items():
            if scmtool_id not in available_scmtools:
                scmtool_choices.append((scmtool_id, name))
                self.scmtool_info[scmtool_id] = {
                    'name': name,
                    'fake': True,
                }

        scmtool_choices.sort(key=lambda x: x[1])
        self.fields['tool'].choices = scmtool_choices

        # Get the current SSH public key that would be used for repositories,
        # if one has been created.
        self.ssh_client = SSHClient(namespace=self.local_site_name)
        ssh_key = self.ssh_client.get_user_key()

        if ssh_key:
            self.public_key = self.ssh_client.get_public_key(ssh_key)
            self.public_key_str = '%s %s' % (
                ssh_key.get_name(),
                ''.join(str(self.public_key).splitlines())
            )
        else:
            self.public_key = None
            self.public_key_str = ''

        # If no SSH key has been created, disable the key association field.
        if not self.public_key:
            self.fields['associate_ssh_key'].help_text = \
                self.NO_KEY_HELP_FMT % local_site_reverse(
                    'settings-ssh',
                    local_site_name=self.local_site_name)
            self.fields['associate_ssh_key'].widget.attrs['disabled'] = \
                'disabled'

        # Set a label for the "public" checkbox that better describes its
        # impact on the repository, given the settings on the server or
        # Local Site.
        if instance:
            instance_local_site = self.local_site or instance.local_site
        else:
            instance_local_site = self.local_site

        if instance_local_site and not instance_local_site.public:
            public_label = (gettext('Accessible to all users on %s')
                            % instance_local_site.name)
            public_help_text = (
                gettext(
                    'Review requests and files on this repository will be '
                    'visible to anyone on %s. Uncheck this box to grant '
                    'access only to specific users and/or to users who are '
                    'members of specific invite-only review groups.')
                % instance_local_site.name)
        elif not instance_local_site or instance_local_site.public:
            siteconfig = SiteConfiguration.objects.get_current()

            if siteconfig.get('auth_require_sitewide_login'):
                public_label = gettext('Accessible to all logged-in users')
                public_help_text = gettext(
                    'Review requests and files on this repository will be '
                    'visible to any logged-in users. Uncheck this box to '
                    'grant access only to specific users and/or to users '
                    'who are members of specific invite-only review groups.')
            else:
                public_label = gettext('Accessible to everyone')
                public_help_text = gettext(
                    'Review requests and files on this repository will be '
                    'visible to any anonymous or logged-in users. Uncheck '
                    'this box to grant access only to specific users and/or '
                    'to users who are members of specific invite-only '
                    'review groups.')

        self.fields['public'].label = public_label
        self.fields['public'].help_text = public_help_text

    @property
    def local_site_name(self):
        """The name of the current Local Site for this form.

        This will be ``None`` if no Local Site is assigned.
        """
        if self.local_site is None:
            return None

        return self.local_site.name

    def iter_subforms(self, bound_only=False, with_auth_forms=False):
        """Iterate through all subforms matching the given criteria.

        This allows callers to easily retrieve all the subforms available
        to the repository form, optionally limiting those to subforms with
        data bound.

        By default, this does not include authentication forms, as those are
        treated specially and should generally not be operated upon in the
        same way as repository and bug tracker subforms.

        The defaults may change, so callers should be explicit about the
        results they want.

        Args:
            bound_only (bool, optional):
                Whether to limit results to bound subforms (those that have
                been populated with data from a form submission).

            with_auth_forms (bool, optional):
                Whether to include authentication forms in the results.

        Yields:
            django.forms.Form:
            Each subform matching the criteria.
        """
        subform_lists = []

        if with_auth_forms:
            subform_lists += [
                self.scmtool_auth_forms.values(),
                self.hosting_auth_forms.values(),
            ]

        subform_lists.append(self.scmtool_repository_forms.values())
        subform_lists += [
            plan_forms.values()
            for plan_forms in chain(
                self.hosting_repository_forms.values(),
                self.hosting_bug_tracker_forms.values())
        ]

        subforms = chain.from_iterable(subform_lists)

        if bound_only:
            subforms = (
                subform
                for subform in subforms
                if subform.is_bound
            )

        return subforms

    def get_repository_already_exists(self):
        """Return whether a repository with these details already exists.

        This will validate the form before returning a result. Callers are
        encouraged to call :py:meth:`is_valid` themselves before calling this.

        Returns:
            bool:
            ``True`` if a repository already exists with this name or path.
            ``False`` if one does not exist.
        """
        if self.is_valid():
            return False

        return (
            Repository.NAME_CONFLICT_ERROR in self.errors.get('name', []) or
            Repository.PATH_CONFLICT_ERROR in self.errors.get('path', [])
        )

    def _get_scmtool_info(self, scmtool_cls):
        """Return the information for a SCMTool.

        Args:
            scmtool_cls (type):
                The SCMTool class, which should be a subclass of
                :py:class:`~reviewboard.scmtools.core.SCMTool`.

        Returns:
            dict:
            Information about the SCMTool.
        """
        info = {}

        for attr in ('name',
                     'supports_pending_changesets',
                     'supports_post_commit'):
            info[attr] = getattr(scmtool_cls, attr)

        return info

    def _get_hosting_service_info(self, hosting_service, hosting_accounts=[],
                                  is_instance_service=False):
        """Return the information for a hosting service.

        Args:
            hosting_service (type):
                The hosting service class.

                This must be a subclass of
                :py:class:`~reviewboard.hostingsvcs.base.hosting_service.
                BaseHostingService`.

            hosting_accounts (list of reviewboard.hostingsvcs.models.
                              HostingServiceAccount, optional):
                A list of the registered
                :py:class:`~reviewboard.hostingsvcs.models.
                HostingServiceAccount`s.

            is_active (boolean, optional):
                Whether this hosting service is currently active, based on
                an existing repository being configured.

        Returns:
            dict:
            Information about the hosting service.
        """
        visible_scmtools = hosting_service.visible_scmtools

        if visible_scmtools is None:
            # visible_scmtools will be None if it just supports all SCMTools.
            # This is a design quirk that works around lack of classproperties,
            # which we'd otherwise use to alias supported_scmtools by default.
            scmtools = hosting_service.supported_scmtools
        elif (is_instance_service and
              self.instance.scmtool_class is not None and
              visible_scmtools != hosting_service.supported_scmtools):
            # Some supported SCMTools aren't shown by default. Want to show
            # only the visible SCMTools, plus whichever one this repository
            # is currently backed by (which likely has only legacy support).
            scmtools = (visible_scmtools +
                        [self.instance.scmtool_class.scmtool_id])
        else:
            scmtools = visible_scmtools

        return {
            'scmtools': sorted(scmtools),
            'plans': [],
            'planInfo': {},
            'self_hosted': hosting_service.self_hosted,
            'needs_authorization': hosting_service.needs_authorization,
            'supports_bug_trackers': hosting_service.supports_bug_trackers,
            'supports_ssh_key_association':
                hosting_service.supports_ssh_key_association,
            'supports_two_factor_auth':
                hosting_service.supports_two_factor_auth,
            'needs_two_factor_auth_code': False,
            'accounts': [
                {
                    'pk': account.pk,
                    'hosting_url': account.hosting_url,
                    'username': account.username,
                    'is_authorized': account.is_authorized,
                }
                for account in hosting_accounts
                if account.service_name == hosting_service.hosting_service_id
            ],
        }

    def _load_scmtool(self, scmtool_cls, is_active):
        """Load forms for a SCMTool.

        This will construct and store the authentication and repository
        information forms. If this is the active SCMTool for the repository,
        then the forms will be loaded with the current data either passed
        to the main form or from the state in the current repository.

        Args:
            scmtool_cls (type):
                The subclass of :py:class:`~reviewboard.scmtools.core.SCMTool`
                that owns these forms.

            is_active (bool):
                Whether this is the active SCMTool for the repository. This
                is only set if working with a plain repository, not one
                backed by a hosting service.
        """
        repository = self.instance
        scmtool_id = scmtool_cls.scmtool_id
        form_kwargs = {
            'local_site': self.local_site,
            'prefix': scmtool_id,
            'repository': repository,
        }

        if is_active:
            initial = model_to_dict(
                repository,
                fields=BaseSCMToolRepositoryForm._MODEL_FIELDS)
            form_kwargs['initial'] = initial

            if self.data:
                data = self.data.copy()

                # We might get form data without prefixes for some fields, such
                # as "path". While the repository page itself will send data
                # with prefixed keys, API consumers and those automating the
                # repository page won't. Look for those keys and convert them
                # to prefixed versions.
                #
                # We'll prioritize any bound form data, and will fall back to
                # initial data if not otherwise found. This ensures we have
                # values for fields like 'path' and 'mirror_path' included.
                for key in self._SCMTOOL_PREFIXLESS_FIELDS:
                    prefixed_key = '%s-%s' % (scmtool_id, key)

                    if not data.get(prefixed_key):
                        if key in data:
                            value = data.pop(key)
                        elif key in initial:
                            value = initial[key]
                        else:
                            continue

                        data[prefixed_key] = value

                form_kwargs['data'] = data

        auth_form = scmtool_cls.create_auth_form(**form_kwargs)
        repo_form = scmtool_cls.create_repository_form(**form_kwargs)

        if is_active:
            auth_form.load()
            repo_form.load()

        # Store these last, in case one of the forms raises an exception.
        # want consistent state.
        self.scmtool_auth_forms[scmtool_id] = auth_form
        self.scmtool_repository_forms[scmtool_id] = repo_form

    def _load_hosting_service(self, hosting_service_id, hosting_service,
                              plan_type_id, plan_type_label, form_class):
        """Load a hosting service form.

        The form will be instantiated and added to the list of forms to be
        rendered, cleaned, loaded, and saved.

        Args:
            hosting_service_id (unicode):
                The ID of the hosting service to load.

            hosting_service (type):
                The hosting service class.

                This will be a subclass of
                :py:class:`~reviewboard.hostingsvcs.base.hosting_service.
                BaseHostingService`.

            plan_type_id (unicode):
                The ID of the hosting plan pertaining to the forms to load.

            plan_type_label (unicode):
                The label shown for the hosting plan.

            form_class (type):
                The hosting service form to use for this plan. This will be
                a subclass of
                :py:class:`~reviewboard.hostingsvcs.forms.HostingServiceForm`.
        """
        repository = self.instance
        plan_info = {}

        if hosting_service.supports_repositories:
            # We only want to load repository data into the form if it's meant
            # for this form. Check the hosting service ID and plan against
            # what's in the submitted form data.
            if (self.data and
                self.data.get('hosting_type') == hosting_service_id and
                (not hosting_service.plans or
                 self.data.get('repository_plan') == plan_type_id)):
                repo_form_data = self.data
            else:
                repo_form_data = None

            form = form_class(data=repo_form_data,
                              repository=repository,
                              hosting_service_cls=hosting_service,
                              local_site=self.local_site)
            self.hosting_repository_forms[hosting_service_id][plan_type_id] = \
                form

            if self.instance:
                form.load(repository)

        if hosting_service.supports_bug_trackers:
            # We only want to load repository data into the form if it's meant
            # for this form. Check the hosting service ID and plan against
            # what's in the submitted form data.
            if (self.data and
                self.data.get('bug_tracker_type') == hosting_service_id and
                not self.data.get('bug_tracker_use_hosting', False) and
                (not hosting_service.plans or
                 self.data.get('bug_tracker_plan') == plan_type_id)):
                bug_tracker_form_data = self.data
            else:
                bug_tracker_form_data = None

            form = form_class(data=bug_tracker_form_data,
                              repository=repository,
                              hosting_service_cls=hosting_service,
                              local_site=self.local_site,
                              prefix='bug_tracker')
            plan_forms = self.hosting_bug_tracker_forms[hosting_service_id]
            plan_forms[plan_type_id] = form

            plan_info['bug_tracker_requires_username'] = \
                hosting_service.get_bug_tracker_requires_username(plan_type_id)

            if self.instance:
                form.load(repository)

        hosting_info = self.hosting_service_info[hosting_service_id]
        hosting_info['planInfo'][plan_type_id] = plan_info
        hosting_info['plans'].append({
            'type': plan_type_id,
            'label': str(plan_type_label),
        })

    def _populate_hosting_service_fields(self):
        """Populates all the main hosting service fields in the form.

        This populates the hosting service type and the repository plan
        on the form. These are only set if operating on an existing
        repository.
        """
        # NOTE: This method *cannot* access anything in the loaded forms or
        #       hosting_service_info attributes.
        hosting_account = self.instance.hosting_account

        if hosting_account:
            hosting_service: BaseHostingService | None

            try:
                hosting_service = hosting_account.service
            except MissingHostingServiceError:
                hosting_service = None

            self.fields['hosting_type'].initial = \
                hosting_account.service_name

            if hosting_service and hosting_service.plans:
                self.fields['repository_plan'].choices = [
                    (plan_id, info['name'])
                    for plan_id, info in hosting_service.plans
                ]

                repository_plan = \
                    self.instance.extra_data.get('repository_plan', None)

                if repository_plan:
                    self.fields['repository_plan'].initial = repository_plan

    def _populate_bug_tracker_fields(self):
        """Populates all the main bug tracker fields in the form.

        This populates the bug tracker type, plan, and other fields
        related to the bug tracker on the form.
        """
        # NOTE: This method *cannot* access anything in the loaded forms or
        #       hosting_service_info attributes.
        data = self.instance.extra_data
        bug_tracker_type = data.get('bug_tracker_type', None)

        if (data.get('bug_tracker_use_hosting', False) and
            self.instance.hosting_account):
            # The user has chosen to use the hosting service's bug tracker. We
            # only care about the checkbox. Don't bother populating the form.
            self.fields['bug_tracker_use_hosting'].initial = True
        elif bug_tracker_type == self.NO_BUG_TRACKER_ID:
            # Do nothing.
            return
        elif (bug_tracker_type is not None and
              bug_tracker_type != self.CUSTOM_BUG_TRACKER_ID):
            # A bug tracker service or custom bug tracker was chosen.
            service = \
                hosting_service_registry.get_hosting_service(bug_tracker_type)

            if not service:
                return

            self.fields['bug_tracker_type'].initial = bug_tracker_type
            self.fields['bug_tracker_hosting_url'].initial = \
                data.get('bug_tracker_hosting_url', None)
            self.fields['bug_tracker_hosting_account_username'].initial = \
                data.get('bug_tracker-hosting_account_username', None)

            if service.plans:
                self.fields['bug_tracker_plan'].choices = [
                    (plan_id, info['name'])
                    for plan_id, info in service.plans
                ]

                self.fields['bug_tracker_plan'].initial = \
                    data.get('bug_tracker_plan', None)
        elif self.instance.bug_tracker:
            # We have a custom bug tracker. There's no point in trying to
            # reverse-match it, because we can potentially be wrong when a
            # hosting service has multiple plans with similar bug tracker
            # URLs, so just show it raw. Admins can migrate it if they want.
            self.fields['bug_tracker_type'].initial = \
                self.CUSTOM_BUG_TRACKER_ID

    def _clean_hosting_info(self):
        """Clean the hosting service information.

        If using a hosting service, this will validate that the data
        provided is valid on that hosting service. Then it will create an
        account and link it, if necessary, with the hosting service.

        Raises:
            django.core.exceptions.ValidationError:
                A field or setting did not pass validation. Details are in
                the error message.
        """
        hosting_type = self.cleaned_data['hosting_type']

        if hosting_type == self.NO_HOSTING_SERVICE_ID:
            self.data['hosting_account'] = None
            self.cleaned_data['hosting_account'] = None
            return

        # This should have been caught during validation, so we can assume
        # it's fine.
        hosting_service_cls = \
            hosting_service_registry.get_hosting_service(hosting_type)
        assert hosting_service_cls

        # Validate that the provided tool is valid for the hosting service.
        tool = self.cleaned_data['tool']
        scmtool_id = tool.scmtool_id

        if (tool.name not in hosting_service_cls.supported_scmtools and
            scmtool_id not in hosting_service_cls.supported_scmtools):
            self.errors['tool'] = self.error_class([
                _('This tool is not supported on the given hosting service')
            ])
            return

        # Get some more information about the hosting service.
        plan = self.cleaned_data['repository_plan'] or self.DEFAULT_PLAN_ID

        # Verify that any hosting account passed in is allowed to work with
        # this type of account.
        hosting_account = self.cleaned_data['hosting_account']

        if (hosting_account and
            (hosting_account.service_name != hosting_type or
             hosting_account.local_site != self.local_site)):
            self.errors['hosting_account'] = self.error_class([
                _('This account is not compatible with this hosting '
                  'service configuration.')
            ])
            return

        repository_form = self.hosting_repository_forms[hosting_type][plan]
        repository_extra_data = repository_form.cleaned_data.copy()

        # If we don't yet have an account, or we have one but it needs to
        # be re-authorized, then we need to go through the entire account
        # updating and authorization process.
        force_authorize = self.cleaned_data['force_authorize']

        if (self.data and
            (not hosting_account or
             not hosting_account.is_authorized or force_authorize)):

            # Rebuild the authentication form, but with data provided to
            # this form, so that we can link or re-authorize an account.
            auth_form = self.hosting_auth_forms[hosting_type]

            auth_form = auth_form.__class__(
                data=self.data,
                prefix=auth_form.prefix,
                hosting_service_cls=auth_form.hosting_service_cls,
                hosting_account=hosting_account,
                repository=self.instance,
                local_site=auth_form.local_site)
            self.hosting_auth_forms[hosting_type] = auth_form

            if not auth_form.is_valid():
                # Copy any errors to the main form, so it'll fail validation
                # and inform the user.
                self.errors.update(auth_form.errors)
                return

            try:
                hosting_account = auth_form.save(
                    extra_authorize_kwargs=repository_extra_data,
                    force_authorize=force_authorize,
                    trust_host=self.cleaned_data['trust_host'])
            except ValueError as e:
                # There was an error with a value provided to the form from
                # The user. Bubble this up.
                self.errors['hosting_account'] = \
                    self.error_class([str(e)])
                return
            except TwoFactorAuthCodeRequiredError as e:
                self.errors['hosting_account'] = \
                    self.error_class([str(e)])
                hosting_info = self.hosting_service_info[hosting_type]
                hosting_info['needs_two_factor_auth_code'] = True
                return
            except AuthorizationError as e:
                self.errors['hosting_account'] = self.error_class([
                    _('Unable to link the account: %s') % e,
                ])
                return
            except (CertificateVerificationError,
                    LegacyUnverifiedCertificateError) as e:
                self.certerror = e
                return
            except Exception as e:
                error = str(e)

                if error.endswith('.'):
                    error = error[:-1]

                self.errors['hosting_account'] = self.error_class([
                    _('Unexpected error when linking the account: %s. '
                      'Additional details may be found in the Review Board '
                      'log file.')
                    % error,
                ])
                return

            # Flag that we've linked the account. If there are any
            # validation errors, and this flag is set, we tell the user
            # that we successfully linked and they don't have to do it
            # again.
            self.hosting_account_linked = True

            # Set this back in the form, so the rest of the form has access.
            self.data['hosting_account'] = hosting_account
            self.cleaned_data['hosting_account'] = hosting_account

        # Set the main repository fields (Path, Mirror Path, etc.) based on
        # the field definitions in the hosting service.
        #
        # This will take into account the hosting service's form data for
        # the given repository plan, the main form data, and the hosting
        # account information.
        #
        # It's expected that the required fields will have validated by now.
        repository_extra_data.update(self.cleaned_data)
        repository_extra_data.update(hosting_account.data)

        try:
            self.subforms_cleaned_data.update(
                hosting_service_cls.get_repository_fields(
                    username=hosting_account.username,
                    hosting_url=hosting_account.hosting_url,
                    plan=plan,
                    tool_name=tool.name,
                    field_vars=repository_extra_data))
        except KeyError as e:
            raise ValidationError([str(e)])

    def _clean_bug_tracker_info(self):
        """Clean the bug tracker information.

        This will figure out the defaults for all the bug tracker fields,
        based on the stored bug tracker settings.
        """
        use_hosting = self.cleaned_data['bug_tracker_use_hosting']
        plan = self.cleaned_data['bug_tracker_plan'] or self.DEFAULT_PLAN_ID
        bug_tracker_type = self.cleaned_data['bug_tracker_type']
        bug_tracker_url = ''

        if use_hosting:
            # We're using the main repository form fields instead of the
            # custom bug tracker fields.
            hosting_type = self.cleaned_data['hosting_type']

            if hosting_type == self.NO_HOSTING_SERVICE_ID:
                self.errors['bug_tracker_use_hosting'] = self.error_class([
                    _('A hosting service must be chosen in order to use this')
                ])
                return

            plan = self.cleaned_data['repository_plan'] or self.DEFAULT_PLAN_ID
            hosting_service_cls = \
                hosting_service_registry.get_hosting_service(hosting_type)

            # We already validated server-side that the hosting service
            # exists.
            assert hosting_service_cls

            if (hosting_service_cls.supports_bug_trackers and
                self.cleaned_data.get('hosting_account')):
                # We have a valid hosting account linked up, so we can
                # process this and copy over the account information.
                form = self.hosting_repository_forms[hosting_type][plan]

                if not form.is_valid():
                    # Skip the rest of this. There's no sense building a URL if
                    # the form's going to display errors.
                    return

                hosting_account = self.cleaned_data['hosting_account']

                new_data = self.cleaned_data.copy()
                new_data.update(form.cleaned_data)
                new_data.update(hosting_account.data)
                new_data['hosting_account_username'] = hosting_account.username
                new_data['hosting_url'] = hosting_account.hosting_url

                try:
                    bug_tracker_url = \
                        hosting_service_cls.get_bug_tracker_field(plan,
                                                                  new_data)
                except KeyError as e:
                    raise ValidationError([str(e)])
        elif bug_tracker_type == self.CUSTOM_BUG_TRACKER_ID:
            # bug_tracker_url should already be in cleaned_data.
            return
        elif bug_tracker_type != self.NO_BUG_TRACKER_ID:
            # We're using a bug tracker of a certain type. We need to
            # get the right data, strip the prefix on the forms, and
            # build the bug tracker URL from that.
            hosting_service_cls = \
                hosting_service_registry.get_hosting_service(bug_tracker_type)

            if not hosting_service_cls:
                self.errors['bug_tracker_type'] = self.error_class([
                    _('This bug tracker type is not supported')
                ])
                return

            form = self.hosting_bug_tracker_forms[bug_tracker_type][plan]

            if not form.is_valid():
                # Skip the rest of this. There's no sense building a URL if
                # the form's going to display errors.
                return

            new_data = dict({
                key: self.cleaned_data['bug_tracker_%s' % key]
                for key in ('hosting_account_username', 'hosting_url')
            }, **{
                # Strip the prefix from each bit of cleaned data in the form.
                key.replace(form.prefix, ''): value
                for key, value in form.cleaned_data.items()
            })

            try:
                bug_tracker_url = hosting_service_cls.get_bug_tracker_field(
                    plan, new_data)
            except KeyError as e:
                raise ValidationError([str(e)])

        self.cleaned_data['bug_tracker'] = bug_tracker_url
        self.data['bug_tracker'] = bug_tracker_url

    def full_clean(self, *args, **kwargs):
        subforms_cleaned_data = {}
        subforms_errors = {}
        required_values = {}

        # Save the required values for all native fields, so that we can
        # restore them we've changed the values and processed forms.
        for field in self.fields.values():
            required_values[field] = field.required

        if self.data:
            hosting_type = self._get_field_data('hosting_type')
            hosting_service = \
                hosting_service_registry.get_hosting_service(hosting_type)
            repository_plan = (self._get_field_data('repository_plan') or
                               self.DEFAULT_PLAN_ID)
            with_auth_forms = (hosting_type == self.NO_HOSTING_SERVICE_ID)

            bug_tracker_use_hosting = \
                self._get_field_data('bug_tracker_use_hosting')

            # If using the hosting service's bug tracker, we want to ignore
            # the bug tracker form (which will be hidden) and just use the
            # hosting service's form.
            if bug_tracker_use_hosting:
                bug_tracker_type = hosting_type
                bug_tracker_service = hosting_service
                bug_tracker_plan = repository_plan
            else:
                bug_tracker_type = self._get_field_data('bug_tracker_type')
                bug_tracker_service = \
                    hosting_service_registry.get_hosting_service(
                        bug_tracker_type)
                bug_tracker_plan = (self._get_field_data('bug_tracker_plan') or
                                    self.DEFAULT_PLAN_ID)

            self.fields['bug_tracker_type'].required = \
                not bug_tracker_use_hosting

            # The repository plan will only be listed if the hosting service
            # lists some plans. Otherwise, there's nothing to require.
            for service, field in ((hosting_service, 'repository_plan'),
                                   (bug_tracker_service, 'bug_tracker_plan')):
                self.fields[field].required = service and service.plans

                if service:
                    self.fields[field].choices = [
                        (id, info['name'])
                        for id, info in service.plans or []
                    ]

            self.fields['bug_tracker_plan'].required = (
                self.fields['bug_tracker_plan'].required and
                not bug_tracker_use_hosting)

            # We want to show this as required (in the label), but not
            # actually require, since we use a blank entry as
            # "Link new account."
            self.fields['hosting_account'].required = False

            # Only require the bug tracker username if the bug tracker field
            # requires the username.
            self.fields['bug_tracker_hosting_account_username'].required = \
                (not bug_tracker_use_hosting and
                 bug_tracker_service and
                 bug_tracker_service.get_bug_tracker_requires_username(
                     bug_tracker_plan))

            # Only require a URL if the bug tracker is self-hosted and
            # we're not using the hosting service's bug tracker.
            self.fields['bug_tracker_hosting_url'].required = (
                not bug_tracker_use_hosting and
                bug_tracker_service and
                bug_tracker_service.self_hosted)
        else:
            with_auth_forms = True

        # Validate the subforms that the repository form is currently working
        # with, and store any data or errors for later.
        #
        # Note that we're not going to validate hosting service authentication
        # forms, which is why we compute with_auth_forms above based on
        # whether a hosting service is selected. We handle those specially in
        # _clean_hosting_info().
        for subform in self.iter_subforms(bound_only=bool(self.data),
                                          with_auth_forms=with_auth_forms):
            if subform.is_valid():
                subforms_cleaned_data.update(subform.cleaned_data)
            else:
                subforms_errors.update(subform.errors)

        self.subforms_cleaned_data = subforms_cleaned_data
        self.subforms_errors = subforms_errors
        self.subforms_valid = not subforms_errors

        super(RepositoryForm, self).full_clean(*args, **kwargs)

        if self.is_valid():
            self.cleaned_data.update(subforms_cleaned_data)
        else:
            self.errors.update(subforms_errors)

        # Undo the required settings above. Now that we're done with them
        # for validation, we want to fix the display so that users don't
        # see the required states change.
        for field, required in required_values.items():
            field.required = required

    def clean(self):
        """Performs validation on the form.

        This will check the form fields for errors, calling out to the
        various clean_* methods.

        It will check the repository path to see if it represents
        a valid repository and if an SSH key or HTTPS certificate needs
        to be verified.

        This will also build repository and bug tracker URLs based on other
        fields set in the form.
        """
        try:
            if not self.errors and self.subforms_valid:
                if not self.limited_to_local_site:
                    try:
                        self.local_site = self.cleaned_data['local_site']
                    except LocalSite.DoesNotExist as e:
                        raise ValidationError(str(e))

                self._clean_hosting_info()
                self._clean_bug_tracker_info()

                # The clean/validation functions could create new errors, so
                # skip validating the repository path if everything else isn't
                # clean. Also skip in the case where the user is hiding the
                # repository.
                if (not self.errors and
                    not self.cleaned_data['reedit_repository'] and
                    self.cleaned_data.get('visible', True)):
                    try:
                        self._verify_repository_path()
                    except ValidationError as e:
                        # We may not be re-raising this exception, which would
                        # cause the exception to be stored in the attribute
                        # in the parent try/except handler. We still want to
                        # store it, so just do that explicitly here.
                        self.form_validation_error = e

                        if e.code == 'cert_unverified':
                            self.certerror = e.params['exception']
                        elif e.code in ('host_key_invalid',
                                        'host_key_unverified'):
                            self.hostkeyerror = e.params['exception']
                        elif e.code == 'missing_ssh_key':
                            self.userkeyerror = e.params['exception']
                        else:
                            raise

                self._clean_ssh_key_association()

                if self.cleaned_data['public']:
                    # Clear out any access lists that may have been set
                    # before. This ensures we don't run into trouble saving
                    # repositories later if a removed user remains in a list.
                    self.cleaned_data.update({
                        'review_groups': [],
                        'users': [],
                    })

            if self.certerror:
                # In the case where there's a certificate error on a hosting
                # service, we'll bail out of the validation process before
                # computing any of the derived fields (like path). This results
                # in the "I trust this host" prompt being shown at the top, but
                # a spurious "Please correct the error below" error shown when
                # no errors are visible. We therefore want to clear out the
                # errors and let the certificate error show on its own. If the
                # user then chooses to trust the cert, the regular verification
                # will run its course.
                self.errors.clear()

            return super(RepositoryForm, self).clean()
        except ValidationError as e:
            # Store this so that the true cause of any ValidationError
            # terminating form cleaning can be looked up. Note that in newer
            # versions of Django, this information is available natively.
            self.form_validation_error = e
            raise

    def _clean_ssh_key_association(self):
        hosting_type = self.cleaned_data['hosting_type']
        hosting_account = self.cleaned_data['hosting_account']

        # Don't proceed if there are already errors, or if not using hosting
        # (hosting type and account should be clean by this point)
        if (self.errors or hosting_type == self.NO_HOSTING_SERVICE_ID or
            not hosting_account):
            return

        hosting_service_cls = \
            hosting_service_registry.get_hosting_service(hosting_type)
        hosting_service = hosting_service_cls(hosting_account)

        # Check the requirements for SSH key association. If the requirements
        # are not met, do not proceed.
        if (not hosting_service_cls.supports_ssh_key_association or
            not self.cleaned_data['associate_ssh_key'] or
            not self.public_key):
            return

        if not self.instance.extra_data:
            # The instance is either a new repository or a repository that
            # was previously configured without a hosting service. In either
            # case, ensure the repository is fully initialized.
            repository = self.save(commit=False)
        else:
            repository = self.instance

        key = self.ssh_client.get_user_key()

        try:
            # Try to upload the key if it hasn't already been associated.
            if not hosting_service.is_ssh_key_associated(repository, key):
                hosting_service.associate_ssh_key(repository, key)
        except SSHKeyAssociationError as e:
            logger.warning('SSHKeyAssociationError for repository "%s" (%s)',
                           repository, str(e))
            raise ValidationError([
                _('Unable to associate SSH key with your hosting service. '
                  'This is most often the result of a problem communicating '
                  'with the hosting service. Please try again later or '
                  'manually upload the SSH key to your hosting service.')
            ])

    def clean_bug_tracker_base_url(self):
        return self.cleaned_data['bug_tracker_base_url'].rstrip('/')

    def clean_bug_tracker_hosting_url(self):
        """Clean the bug tracker hosting URL.

        This will strip all whitespace from the URL.

        Returns:
            unicode:
            The hosting URL with whitespace stripped.
        """
        return self.cleaned_data['bug_tracker_hosting_url'].strip()

    def clean_hosting_type(self):
        """Validates that the hosting type represents a valid hosting service.

        This won't do anything if no hosting service is used.
        """
        hosting_type = self.cleaned_data['hosting_type']

        if hosting_type != self.NO_HOSTING_SERVICE_ID:
            hosting_service = \
                hosting_service_registry.get_hosting_service(hosting_type)

            if not hosting_service:
                raise ValidationError([_('Not a valid hosting service')])

        return hosting_type

    def clean_bug_tracker_type(self):
        """Validates that the bug tracker type represents a valid hosting
        service.

        This won't do anything if no hosting service is used.
        """
        bug_tracker_type = (self.cleaned_data['bug_tracker_type'] or
                            self.NO_BUG_TRACKER_ID)

        if bug_tracker_type not in self.IGNORED_SERVICE_IDS:
            hosting_service = \
                hosting_service_registry.get_hosting_service(bug_tracker_type)

            if (not hosting_service or
                not hosting_service.supports_bug_trackers):
                raise ValidationError([_('Not a valid hosting service')])

        return bug_tracker_type

    def clean_tool(self):
        """Check the SCMTool used for this repository.

        This will ensure the selected SCMTool is valid and that its
        dependencies all exist.

        Returns:
            reviewboard.scmtools.models.Tool:
            The Tool model entry to assign to the repository.

        Raises:
            django.core.exceptions.ValidationError:
                The tool was invalid, or one of its dependencies was missing.
        """
        errors = []
        scmtool_id = self.cleaned_data['tool']

        try:
            scmtool = scmtools_registry.get_by_id(scmtool_id)
        except KeyError:
            raise ValidationError(['Invalid SCMTool.'])

        for dep in scmtool.dependencies.get('modules', []):
            if not has_module(dep):
                errors.append(_('The Python module "%s" is not installed. '
                                'You may need to restart the server '
                                'after installing it.') % dep)

        for dep in scmtool.dependencies.get('executables', []):
            if not is_exe_in_path(dep):
                if sys.platform == 'win32':
                    exe_name = '%s.exe' % dep
                else:
                    exe_name = dep

                errors.append(_('The executable "%s" is not in the path.')
                              % exe_name)

        if errors:
            raise ValidationError(errors)

        return scmtool

    def clean_extra_data(self):
        """Clean the extra_data field.

        This will ensure that the field is always a dictionary.

        Returns:
            dict:
            The extra_data dictionary.

        Raises:
            django.core.exceptions.ValidationError:
                The value was not a dictionary.
        """
        extra_data = self.cleaned_data['extra_data'] or {}

        if not isinstance(extra_data, dict):
            raise ValidationError(gettext(
                'This must be a JSON object/dictionary.'))

        return extra_data

    def is_valid(self):
        """Return whether or not the form is valid.

        This will return True if the form fields are all valid, if there's
        no certificate error, host key error, and if the form isn't
        being re-displayed after canceling an SSH key or HTTPS certificate
        verification.

        This also takes into account the validity of any relevant subforms.

        Returns:
            bool:
            ``True`` if the form is valid. ``False`` if it is not.
        """
        return (super(RepositoryForm, self).is_valid() and
                not self.hostkeyerror and
                not self.certerror and
                not self.userkeyerror and
                not self.cleaned_data['reedit_repository'] and
                self.subforms_valid)

    def save(self, commit=True):
        """Save the repository.

        This will save some of the general information for the repository and
        the hosting service (if selected), and use the subforms to save the
        rest.

        This must be called after :py:meth:`is_valid`.

        Args:
            commit (bool, optional):
                Whether to save the repository to the database.

                If ``False``, the repository will be constructed but not saved.
                It is then the responsibility of the caller to call
                :py:meth:`Repository.save()
                <reviewboard.scmtools.models.Repository.save>` and
                :py:meth:`save_m2m`.

        Returns:
            reviewboard.scmtools.models.Repository:
            The resulting repository.

        Raises:
            ValueError:
                The form had pending errors, and could not be saved.
        """
        # Before we make any changes, we want to see if the hosting service
        # or SCMTool have changed. These will determine whether we need to
        # clear out some old extra_data state.
        old_hosting_service = self.instance.hosting_service
        old_scmtool_cls = self.instance.scmtool_class

        extra_data = self.cleaned_data['extra_data']
        hosting_type = self.cleaned_data['hosting_type']
        scmtool = self.cleaned_data['tool']
        scmtool_id = scmtool.scmtool_id
        repository_plan = self.cleaned_data.get('repository_plan')
        bug_tracker_plan = self.cleaned_data.get('bug_tracker_plan')

        extra_data_prefixes_to_remove = ['bug_tracker_', 'bug_tracker-']

        if old_hosting_service is not None:
            old_hosting_type = old_hosting_service.hosting_service_id
            old_repository_plan = \
                self.instance.extra_data.get('repository_plan')

            if (hosting_type != old_hosting_type or
                repository_plan != old_repository_plan):
                extra_data_prefixes_to_remove.append('%s_' % old_hosting_type)

        if old_scmtool_cls is not None:
            old_scmtool_id = old_scmtool_cls.scmtool_id

            if old_scmtool_id != scmtool_id:
                extra_data_prefixes_to_remove.append('%s_' % old_scmtool_id)

        # Start removing the keys we don't want.
        for key in ('cert',
                    'hosting_url',
                    'repository_plan',
                    'use_ticket_auth'):
            extra_data.pop(key, None)

        extra_data_prefixes_to_remove = tuple(extra_data_prefixes_to_remove)

        if extra_data_prefixes_to_remove:
            for key in list(extra_data.keys()):
                if key.startswith(extra_data_prefixes_to_remove):
                    del extra_data[key]

        # We can now start populating the repository's fields.
        repository = super(RepositoryForm, self).save(commit=False)
        repository.tool = Tool.objects.get(class_name=scmtool.class_name)
        repository.scmtool_id = scmtool.scmtool_id
        repository.path = self.cleaned_data['path']
        repository.mirror_path = self.cleaned_data.get('mirror_path', '')
        repository.raw_file_url = self.cleaned_data.get('raw_file_url', '')

        bug_tracker_use_hosting = self.cleaned_data['bug_tracker_use_hosting']

        if hosting_type == self.NO_HOSTING_SERVICE_NAME:
            service = None
        else:
            service = \
                hosting_service_registry.get_hosting_service(hosting_type)

        if service is not None:
            repository.username = ''
            repository.password = ''

            repository.extra_data.update({
                'repository_plan': repository_plan,
                'bug_tracker_use_hosting': bug_tracker_use_hosting,
            })

            if service.self_hosted:
                repository.extra_data['hosting_url'] = \
                    repository.hosting_account.hosting_url

            if hosting_type in self.hosting_repository_forms:
                plan = repository_plan or self.DEFAULT_PLAN_ID
                self.hosting_repository_forms[hosting_type][plan].save(
                    repository)
        else:
            self.scmtool_auth_forms[scmtool_id].save()
            self.scmtool_repository_forms[scmtool_id].save()

        if self.cert:
            repository.extra_data['cert'] = self.cert

        if not bug_tracker_use_hosting:
            bug_tracker_type = self.cleaned_data['bug_tracker_type']

            if bug_tracker_type in self.hosting_bug_tracker_forms:
                plan = bug_tracker_plan or self.DEFAULT_PLAN_ID
                self.hosting_bug_tracker_forms[bug_tracker_type][plan].save(
                    repository)
                repository.extra_data.update({
                    'bug_tracker_type': bug_tracker_type,
                    'bug_tracker_plan': plan,
                })

                bug_tracker_service = \
                    hosting_service_registry.get_hosting_service(
                        bug_tracker_type)
                assert bug_tracker_service

                if bug_tracker_service.self_hosted:
                    repository.extra_data['bug_tracker_hosting_url'] = \
                        self.cleaned_data['bug_tracker_hosting_url']

                if bug_tracker_service.get_bug_tracker_requires_username(plan):
                    repository.extra_data.update({
                        'bug_tracker-hosting_account_username':
                            self.cleaned_data[
                                'bug_tracker_hosting_account_username'],
                    })

        if commit:
            repository.save()
            self.save_m2m()

        return repository

    def _verify_repository_path(self):
        """Verify the repository path to check if it's valid.

        This will check if the repository exists and is accessible, and
        confirm whether the SSH key or HTTPS certificate needs to be manually
        verified by the administrator.

        Raises:
            django.core.exceptions.ValidationError:
                The repository information fails to pass validation. Details
                and explicit error codes will be in the exception.
        """
        scmtool = self.cleaned_data.get('tool')

        if not scmtool:
            # This failed validation earlier, so bail.
            return

        subforms_cleaned_data = self.subforms_cleaned_data
        assert subforms_cleaned_data is not None
        path = subforms_cleaned_data.get('path')
        username = subforms_cleaned_data.get('username')
        password = subforms_cleaned_data.get('password')
        hosting_type = self.cleaned_data['hosting_type']
        hosting_service_cls = \
            hosting_service_registry.get_hosting_service(hosting_type)
        hosting_service = None
        plan = None

        if hosting_service_cls:
            hosting_service = hosting_service_cls(
                self.cleaned_data['hosting_account'])

            if hosting_service:
                plan = (self.cleaned_data['repository_plan'] or
                        self.DEFAULT_PLAN_ID)
        else:
            # For plain repositories, a SCMTool can specify that it prefers
            # using a Mirror Path instead of a Path for all communication and
            # repository validation. This is available for historical reasons.
            # We'll check for that only if we're not using a hosting service
            # (which should have its own custom repository checks).
            if scmtool.prefers_mirror_path:
                path = subforms_cleaned_data.get('mirror_path') or path

        if not path:
            # This may have been caught during form validation, but it depends
            # on the subform, so check again.
            self._errors['path'] = self.error_class([
                gettext('The repository path cannot be empty')
            ])
            return

        # Build the data to pass to check_repository().
        #
        # For backwards-compatibility, we'll ensure we always have some
        # minimum set of fields.
        repository_extra_data = {
            'username': None,
            'password': None,
            'path': None,
        }
        repository_extra_data.update(subforms_cleaned_data)

        local_site_name = self.local_site_name

        while 1:
            # Keep doing this until we have an error we don't want
            # to ignore, or it's successful.
            try:
                if hosting_service:
                    hosting_service.check_repository(
                        scmtool_class=scmtool,
                        tool_name=scmtool.name,
                        local_site=self.local_site,
                        local_site_name=local_site_name,
                        plan=plan,
                        **repository_extra_data)
                else:
                    scmtool.check_repository(
                        local_site=self.local_site,
                        local_site_name=local_site_name,
                        **repository_extra_data)

                # Success.
                break
            except RepositoryNotFoundError as e:
                raise ValidationError(str(e),
                                      code='repository_not_found')
            except BadHostKeyError as e:
                if not self.cleaned_data['trust_host']:
                    raise ValidationError(
                        str(e),
                        code='host_key_invalid',
                        params={
                            'exception': e,
                        })

                try:
                    self.ssh_client.replace_host_key(e.hostname,
                                                     e.raw_expected_key,
                                                     e.raw_key)
                except IOError as e:
                    raise ValidationError(str(e),
                                          code='replace_host_key_failed')
            except UnknownHostKeyError as e:
                if not self.cleaned_data['trust_host']:
                    raise ValidationError(
                        str(e),
                        code='host_key_unverified',
                        params={
                            'exception': e,
                        })

                try:
                    self.ssh_client.add_host_key(e.hostname, e.raw_key)
                except IOError as e:
                    raise ValidationError(str(e), code='add_host_key_failed')
            except (CertificateVerificationError,
                    LegacyUnverifiedCertificateError) as e:
                if not self.cleaned_data['trust_host']:
                    raise ValidationError(
                        str(e),
                        code='cert_unverified',
                        params={
                            'exception': e,
                        })

                try:
                    self.cert = scmtool.accept_certificate(
                        path,
                        username=username,
                        password=password,
                        local_site_name=local_site_name,
                        certificate=e.certificate)
                except IOError as e:
                    raise ValidationError(str(e), code='accept_cert_failed')
            except AuthenticationError as e:
                if 'publickey' in e.allowed_types and e.user_key is None:
                    raise ValidationError(
                        str(e),
                        code='missing_ssh_key',
                        params={
                            'exception': e,
                        })

                raise ValidationError(str(e),
                                      code='repo_auth_failed')
            except Exception as e:
                logger.exception(
                    'Unexpected exception while verifying repository path for '
                    'hosting service %r using plan %r and tool %r: %s',
                    hosting_service, plan, scmtool, e)

                try:
                    text = str(e)
                except UnicodeDecodeError:
                    text = str(e, 'ascii', 'replace')

                if isinstance(e, HostingServiceError):
                    code = 'unexpected_hosting_service_failure'
                elif isinstance(e, SSHError):
                    code = 'unexpected_ssh_failure'
                elif isinstance(e, SCMError):
                    code = 'unexpected_scm_failure'
                else:
                    code = 'unexpected_failure'

                if getattr(e, 'help_link', None):
                    text = format_html(_('{0} <a href="{1}">{2}</a>'),
                                       text, e.help_link,
                                       e.help_link_text)

                raise ValidationError(text, code=code)

    def _get_field_data(self, field):
        return self[field].data or self.fields[field].initial

    class Meta:
        model = Repository
        widgets = {
            'bug_tracker': forms.TextInput(attrs={'size': '60'}),
            'name': forms.TextInput(attrs={'size': '30',
                                           'autocomplete': 'off'}),
            'review_groups': FilteredSelectMultiple(
                _('review groups with access'), False),
        }
        fields = '__all__'
        exclude = ('username', 'password', 'path', 'mirror_path',
                   'raw_file_url', 'tool')



"""Certificate storage backend registry.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import Iterator, Optional, Type

from django.utils.translation import gettext_lazy as _
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         NOT_REGISTERED)

from reviewboard.registries.registry import Registry
from reviewboard.certs.storage.base import BaseCertificateStorageBackend


class CertificateStorageBackendRegistry(
    Registry[Type[BaseCertificateStorageBackend]]
):
    """Registry for managing certificate storage backends.

    By default, this includes the file-based storage backend. Extensions can
    register additional backends, which can then be activated for certificate
    storage management.

    Version Added:
        6.0
    """

    lookup_attrs = ('backend_id',)

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered certificate storage backend.'
        ),
        NOT_REGISTERED: _(
            '"%(attr_value)s" is not a registered certificate storage backend.'
        ),
    }

    def get_backend(
        self,
        backend_id: str,
    ) -> Optional[Type[BaseCertificateStorageBackend]]:
        """Return a storage backend class with the given ID.

        Args:
            backend_id (str):
                The ID of the storage backend.

        Returns:
            type:
            The storage backend class, or ``None`` if not found.
        """
        return self.get('backend_id', backend_id)

    def get_defaults(self) -> Iterator[Type[BaseCertificateStorageBackend]]:
        """Return the default storage backend classes.

        Yields:
            type:
            Each default storage backend class.
        """
        from reviewboard.certs.storage.file_storage import \
            FileCertificateStorageBackend

        yield FileCertificateStorageBackend



"""Unit tests for reviewboard.cmdline.utils.console."""

import io
import math
import shutil

import kgb
from django.core.management.color import supports_color

from reviewboard.cmdline.utils.console import Console
from reviewboard.testing.testcase import TestCase


class ConsoleTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.cmdline.utils.console.Console."""

    def test_with_non_utf8_streams(self):
        """Testing Console with non-utf-8 stdout/stderr streams"""
        stdout_buffer = io.BytesIO()
        stdout = io.TextIOWrapper(stdout_buffer, encoding='latin1')

        stderr_buffer = io.BytesIO()
        stderr = io.TextIOWrapper(stderr_buffer, encoding='latin1')

        try:
            console = Console(stdout=stdout,
                              stderr=stderr)

            # This will output Unicode content to stdout, and should fail
            # if there's an encoding issue.
            console.print('\U0001f9f8')

            # There's no wrapper for stderr, so write to it directly.
            console.stderr.write('\U0001f534')

            # Make sure we got the results we expected.
            self.assertEqual(stdout_buffer.getvalue(),
                             b'\xf0\x9f\xa7\xb8\n')
            self.assertEqual(stderr_buffer.getvalue(),
                             b'\xf0\x9f\x94\xb4')
        finally:
            stdout.close()
            stderr.close()

    def test_print_with_styled_prefix(self):
        """Testing Console.print with styled indent prefix"""
        self.spy_on(supports_color, op=kgb.SpyOpReturn(True))
        self.spy_on(shutil.get_terminal_size, op=kgb.SpyOpReturn((50, 40)))

        stdout_buffer = io.BytesIO()
        stdout = io.TextIOWrapper(stdout_buffer)

        try:
            console = Console(stdout=stdout,
                              allow_color=True)

            prefix_len = len('Warning: ')
            avail_len = console.term_width - prefix_len - 1

            console.warning('x ' * math.ceil(3 * (avail_len / 2)))

            self.assertEqual(
                stdout_buffer.getvalue(),
                b'\n'
                b'\x1b[33;1mWarning: \x1b[0m'
                b'x x x x x x x x x x x x x x x x x x x x x\n'
                b'         x x x x x x x x x x x x x x x x x x x x x\n'
                b'         x x x x x x x x x x x x x x x x x x\n'
                b'\n')
        finally:
            stdout.close()

    def test_print_with_prefix_and_multiple_paragraphs(self):
        """Testing Console.print with indent prefix and multiple paragraphs"""
        self.spy_on(shutil.get_terminal_size, op=kgb.SpyOpReturn((50, 40)))

        stdout_buffer = io.BytesIO()
        stdout = io.TextIOWrapper(stdout_buffer)

        try:
            console = Console(stdout=stdout)
            console.warning('line 1\n'
                            'line 2\n'
                            'line 3\n')

            self.assertEqual(
                stdout_buffer.getvalue(),
                b'\n'
                b'Warning: line 1\n'
                b'\n'
                b'         line 2\n'
                b'\n'
                b'         line 3\n'
                b'\n')
        finally:
            stdout.close()



"""Change ReviewRequest.repository to allow NULL.

Version Added:
    1.5
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('ReviewRequest', 'repository', initial=None, null=True)
]



"""FileDiff model definition."""

from __future__ import annotations

import logging
from itertools import chain
from typing import ClassVar, Optional, TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import Base64Field, JSONField

from reviewboard.diffviewer.managers import FileDiffManager
from reviewboard.diffviewer.models.diffcommit import DiffCommit
from reviewboard.diffviewer.models.legacy_file_diff_data import \
    LegacyFileDiffData
from reviewboard.diffviewer.models.raw_file_diff_data import RawFileDiffData
from reviewboard.scmtools.core import PRE_CREATION

if TYPE_CHECKING:
    from reviewboard.reviews.models import ReviewRequest
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class FileDiff(models.Model):
    """A diff of a single file.

    This contains the patch and information needed to produce original and
    patched versions of a single file in a repository.
    """

    _ANCESTORS_KEY = '__ancestors'

    COPIED = 'C'
    DELETED = 'D'
    MODIFIED = 'M'
    MOVED = 'V'

    STATUSES = (
        (COPIED, _('Copied')),
        (DELETED, _('Deleted')),
        (MODIFIED, _('Modified')),
        (MOVED, _('Moved')),
    )

    _IS_PARENT_EMPTY_KEY = '__parent_diff_empty'

    diffset = models.ForeignKey('DiffSet',
                                on_delete=models.CASCADE,
                                related_name='files',
                                verbose_name=_('diff set'))

    commit = models.ForeignKey(DiffCommit,
                               on_delete=models.CASCADE,
                               related_name='files',
                               verbose_name=_('diff commit'),
                               null=True)

    source_file = models.CharField(_('source file'), max_length=1024)
    dest_file = models.CharField(_('destination file'), max_length=1024)
    source_revision = models.CharField(_('source file revision'),
                                       max_length=512)
    dest_detail = models.CharField(_('destination file details'),
                                   max_length=512)
    binary = models.BooleanField(_('binary file'), default=False)
    status = models.CharField(_('status'), max_length=1, choices=STATUSES)

    diff64 = Base64Field(
        _('diff'),
        db_column='diff_base64',
        blank=True)
    legacy_diff_hash = models.ForeignKey(
        LegacyFileDiffData,
        on_delete=models.SET_NULL,
        db_column='diff_hash_id',
        related_name='filediffs',
        null=True,
        blank=True)
    diff_hash = models.ForeignKey(
        RawFileDiffData,
        on_delete=models.SET_NULL,
        db_column='raw_diff_hash_id',
        related_name='filediffs',
        null=True,
        blank=True)

    parent_diff64 = Base64Field(
        _('parent diff'),
        db_column='parent_diff_base64',
        blank=True)
    legacy_parent_diff_hash = models.ForeignKey(
        LegacyFileDiffData,
        on_delete=models.SET_NULL,
        db_column='parent_diff_hash_id',
        related_name='parent_filediffs',
        null=True,
        blank=True)
    parent_diff_hash = models.ForeignKey(
        RawFileDiffData,
        on_delete=models.SET_NULL,
        db_column='raw_parent_diff_hash_id',
        related_name='parent_filediffs',
        null=True,
        blank=True)

    extra_data = JSONField(null=True)

    objects: ClassVar[FileDiffManager] = FileDiffManager()

    @property
    def source_file_display(self):
        """The displayed filename for the source/original file.

        This may be different than :py:attr:`source_file`, as the associated
        :py:class:`~reviewboard.scmtools.core.SCMTool` may normalize it for
        display.

        Type:
            unicode
        """
        tool = self.diffset.repository.get_scmtool()
        return tool.normalize_path_for_display(
            self.source_file,
            extra_data=self.extra_data)

    @property
    def dest_file_display(self):
        """The displayed filename for the destination/modified file.

        This may be different than :py:attr:`dest_file`, as the associated
        :py:class:`~reviewboard.scmtools.core.SCMTool` may normalize it for
        display.

        Type:
            unicode
        """
        tool = self.diffset.repository.get_scmtool()
        return tool.normalize_path_for_display(
            self.dest_file,
            extra_data=self.extra_data)

    @property
    def deleted(self):
        return self.status == self.DELETED

    @property
    def copied(self):
        return self.status == self.COPIED

    @property
    def moved(self):
        return self.status == self.MOVED

    @property
    def modified(self):
        """Whether this file is a modification to an existing file."""
        return self.status == self.MODIFIED

    @property
    def is_new(self):
        return self.source_revision == PRE_CREATION

    @property
    def is_symlink(self):
        """Whether this file represents a symlink.

        This represents the value of the ``is_symlink`` key in
        :py:attr:`extra_data`.

        Version Added:
            4.0.6

        Type:
            bool
        """
        return self.extra_data.get('is_symlink', False)

    @is_symlink.setter
    def is_symlink(self, value):
        """Set whether this file represents a symlink.

        This will set the ``is_symlink`` key in :py:attr:`extra_data` if
        setting a symlink.

        Version Added:
            4.0.6

        Args:
            value (bool):
                Whether this represents a symlink.
        """
        self.extra_data['is_symlink'] = value

    @property
    def old_symlink_target(self):
        """The old target for the symlink.

        This represents the value of the ``old_symlink_target`` key in
        :py:attr:`extra_data`.

        Version Added:
            4.0.6

        Type:
            unicode
        """
        return self.extra_data.get('old_symlink_target')

    @old_symlink_target.setter
    def old_symlink_target(self, value):
        """Set whether this file represents a symlink.

        This will set the ``old_symlink_target`` key in :py:attr:`extra_data`.

        Version Added:
            4.0.6

        Args:
            value (unicode):
                The old target for the symlink.
        """
        self.extra_data['old_symlink_target'] = value

    @property
    def new_symlink_target(self):
        """The new target for the symlink.

        This represents the value of the ``new_symlink_target`` key in
        :py:attr:`extra_data`.

        Version Added:
            4.0.6

        Type:
            unicode
        """
        return self.extra_data.get('new_symlink_target')

    @new_symlink_target.setter
    def new_symlink_target(self, value):
        """Set whether this file represents a symlink.

        This will set the ``new_symlink_target`` key in :py:attr:`extra_data`.

        Version Added:
            4.0.6

        Args:
            value (unicode):
                The new target for the symlink.
        """
        self.extra_data['new_symlink_target'] = value

    @property
    def old_unix_mode(self):
        """The old UNIX permissions mode for the file.

        This represents the value of the ``old_unix_mode`` key in
        :py:attr:`extra_data`. It is stored as a string representing an
        octal.

        Version Added:
            4.0.6

        Type:
            unicode
        """
        return self.extra_data.get('old_unix_mode')

    @old_unix_mode.setter
    def old_unix_mode(self, value):
        """Set the old UNIX permissions mode for the file.

        This will set the ``old_unix_mode`` key in :py:attr:`extra_data`
        to a string-encoded UNIX file mode.

        If setting to ``None``, this will delete the key.

        Version Added:
            4.0.6

        Args:
            value (unicode):
                The old UNIX mode.
        """
        if value is None:
            self.extra_data.pop('old_unix_mode', None)
        else:
            self.extra_data['old_unix_mode'] = value

    @property
    def new_unix_mode(self):
        """The new UNIX permissions mode for the file.

        This represents the value of the ``new_unix_mode`` key in
        :py:attr:`extra_data`. It is stored as a string representing an
        octal.

        Version Added:
            4.0.6

        Type:
            unicode
        """
        return self.extra_data.get('new_unix_mode')

    @new_unix_mode.setter
    def new_unix_mode(self, value):
        """Set the new UNIX permissions mode for the file.

        This will set the ``new_unix_mode`` key in :py:attr:`extra_data`
        to a string-encoded UNIX file mode.

        If setting to ``None``, this will delete the key.

        Version Added:
            4.0.6

        Args:
            value (unicode):
                The old UNIX mode.
        """
        if value is None:
            self.extra_data.pop('new_unix_mode', None)
        else:
            self.extra_data['new_unix_mode'] = value

    @property
    def status_string(self):
        """The FileDiff's status as a human-readable string."""
        if self.status == FileDiff.COPIED:
            return 'copied'
        elif self.status == FileDiff.DELETED:
            return 'deleted'
        elif self.status == FileDiff.MODIFIED:
            return 'modified'
        elif self.status == FileDiff.MOVED:
            return 'moved'
        else:
            logger.error('Unknown FileDiff status %r for FileDiff %s',
                         self.status, self.pk)
            return 'unknown'

    def _get_diff(self):
        if self._needs_diff_migration():
            self._migrate_diff_data()

        return self.diff_hash.content

    def _set_diff(self, diff):
        # Add hash to table if it doesn't exist, and set diff_hash to this.
        self.diff_hash, is_new = \
            RawFileDiffData.objects.get_or_create_from_data(diff)
        self.diff64 = b''

        return is_new

    diff = property(_get_diff, _set_diff)

    @property
    def is_diff_empty(self):
        """Whether or not the diff is empty."""
        line_counts = self.get_line_counts()

        return (line_counts['raw_insert_count'] == 0 and
                line_counts['raw_delete_count'] == 0)

    def _get_parent_diff(self):
        if self._needs_parent_diff_migration():
            self._migrate_diff_data()

        if self.parent_diff_hash:
            return self.parent_diff_hash.content
        else:
            return None

    def _set_parent_diff(self, parent_diff):
        if not parent_diff:
            return False

        # Add hash to table if it doesn't exist, and set diff_hash to this.
        self.parent_diff_hash, is_new = \
            RawFileDiffData.objects.get_or_create_from_data(parent_diff)
        self.parent_diff64 = b''

        return is_new

    parent_diff = property(_get_parent_diff, _set_parent_diff)

    def is_parent_diff_empty(self, cache_only=False):
        """Return whether or not the parent diff is empty.

        Args:
            cache_only (bool, optional):
                Whether or not to only use cached results.

        Returns:
            bool:
            Whether or not the parent diff is empty. This is true if either
            there is no parent diff or if the parent diff has no insertions and
            no deletions.
        """
        assert self.parent_diff_hash_id is not None, 'No parent diff.'

        if cache_only:
            return self.extra_data.get(self._IS_PARENT_EMPTY_KEY, False)

        if (not self.extra_data or
            self._IS_PARENT_EMPTY_KEY not in self.extra_data):
            parent_diff_hash = self.parent_diff_hash

            if (parent_diff_hash.insert_count is None or
                parent_diff_hash.delete_count is None):
                tool = self.get_repository().get_scmtool()
                parent_diff_hash.recalculate_line_counts(tool)

            self.extra_data[self._IS_PARENT_EMPTY_KEY] = (
                parent_diff_hash.insert_count == 0 and
                parent_diff_hash.delete_count == 0)

            self.save(update_fields=('extra_data',))

        return self.extra_data[self._IS_PARENT_EMPTY_KEY]

    @property
    def orig_sha1(self):
        """The computed SHA1 hash of the original file.

        This may be ``None``, in which case it will be populated when the
        diff is next viewed.
        """
        return self.extra_data.get('orig_sha1')

    @property
    def patched_sha1(self):
        """The computed SHA1 hash of the patched file.

        This may be ``None``, in which case it will be populated when the
        diff is next viewed.
        """
        return self.extra_data.get('patched_sha1')

    @property
    def orig_sha256(self):
        """The computed SHA256 hash of the original file.

        This may be ``None``, in which case it will be populated when the
        diff is next viewed.

        Version Added:
            4.0
        """
        # Avoid a circular import.
        from reviewboard.attachments.models import FileAttachment

        if not self.binary:
            return self.extra_data.get('orig_sha256')
        else:
            attachment = FileAttachment.objects.get_for_filediff(
                self,
                modified=False)

            if attachment:
                return attachment.sha256_checksum
            else:
                return None

    @property
    def patched_sha256(self):
        """The computed SHA256 hash of the patched file.

        This may be ``None``, in which case it will be populated when the
        diff is next viewed.

        Version Added:
            4.0
        """
        # Avoid a circular import.
        from reviewboard.attachments.models import FileAttachment

        if not self.binary:
            return self.extra_data.get('patched_sha256')
        else:
            attachment = FileAttachment.objects.get_for_filediff(self)

            if attachment:
                return attachment.sha256_checksum
            else:
                return None

    @property
    def encoding(self):
        """The encoding of the source and patched file.

        This will check the ``encoding`` key in :py:attr:`extra_data`.
        If not available, then this will be ``None``.

        Version Added:
            4.0
        """
        return self.extra_data.get('encoding')

    def get_line_counts(self):
        """Return the stored line counts for the diff.

        This will return all the types of line counts that can be set.

        Returns:
            dict:
            A dictionary with the following keys:

            * ``raw_insert_count``
            * ``raw_delete_count``
            * ``insert_count``
            * ``delete_count``
            * ``replace_count``
            * ``equal_count``
            * ``total_line_count``

            These are not all guaranteed to have values set, and may instead be
            ``None``. Only ``raw_insert_count``, ``raw_delete_count``
            ``insert_count``, and ``delete_count`` are guaranteed to have
            values set.

            If there isn't a processed number of inserts or deletes stored,
            then ``insert_count`` and ``delete_count`` will be equal to the raw
            versions.
        """
        if self.binary:
            if self.is_new:
                return {
                    'delete_count': 0,
                    'equal_count': 0,
                    'insert_count': 1,
                    'raw_delete_count': 0,
                    'raw_insert_count': 0,
                    'replace_count': 0,
                    'total_line_count': 1,
                }
            elif self.deleted:
                return {
                    'delete_count': 1,
                    'equal_count': 0,
                    'insert_count': 0,
                    'raw_delete_count': 0,
                    'raw_insert_count': 0,
                    'replace_count': 0,
                    'total_line_count': 1,
                }
            else:
                return {
                    'delete_count': 0,
                    'equal_count': 0,
                    'insert_count': 0,
                    'raw_delete_count': 0,
                    'raw_insert_count': 0,
                    'replace_count': 1,
                    'total_line_count': 1,
                }

        if ('raw_insert_count' not in self.extra_data or
            'raw_delete_count' not in self.extra_data):
            if not self.diff_hash:
                self._migrate_diff_data()

            if self.diff_hash.insert_count is None:
                self._recalculate_line_counts(self.diff_hash)

            self.extra_data.update({
                'raw_insert_count': self.diff_hash.insert_count,
                'raw_delete_count': self.diff_hash.delete_count,
            })

            if self.pk:
                self.save(update_fields=['extra_data'])

        raw_insert_count = self.extra_data['raw_insert_count']
        raw_delete_count = self.extra_data['raw_delete_count']

        return {
            'raw_insert_count': raw_insert_count,
            'raw_delete_count': raw_delete_count,
            'insert_count': self.extra_data.get('insert_count',
                                                raw_insert_count),
            'delete_count': self.extra_data.get('delete_count',
                                                raw_delete_count),
            'replace_count': self.extra_data.get('replace_count'),
            'equal_count': self.extra_data.get('equal_count'),
            'total_line_count': self.extra_data.get('total_line_count'),
        }

    def set_line_counts(self, raw_insert_count=None, raw_delete_count=None,
                        insert_count=None, delete_count=None,
                        replace_count=None, equal_count=None,
                        total_line_count=None):
        """Set the line counts on the FileDiff.

        There are many types of useful line counts that can be set.

        Args:
            raw_insert_count (int, optional):
                The insert count on the original patch.

                This will be set on the
                :py:class:`reviewboard.diffviewer.models.raw_file_diff_data.RawFileDiffData`
                as well.

            raw_delete_count (int, optional):
                The delete count in the original patch.

                This will be set on the
                :py:class:`reviewboard.diffviewer.models.raw_file_diff_data.RawFileDiffData`
                as well.

            insert_count (int, optional):
                The number of lines that were inserted in the diff.

            delete_count (int, optional):
                The number of lines that were deleted in the diff.

            replace_count (int, optional):
                The number of lines that were replaced in the diff.

            equal_count (int, optional):
                The number of lines that were identical in the diff.

            total_line_count (int, optional):
                The total line count.
        """
        updated = False

        if not self.diff_hash_id:
            # This really shouldn't happen, but if it does, we should handle
            # it gracefully.
            logger.warning('Attempting to call set_line_counts on '
                           'un-migrated FileDiff %s',
                           self.pk)
            self._migrate_diff_data(False)

        if (insert_count is not None and
            raw_insert_count is not None and
            self.diff_hash.insert_count is not None and
            self.diff_hash.insert_count != insert_count):
            # Allow overriding, but warn. This really shouldn't be called.
            logger.warning('Attempting to override insert count on '
                           'RawFileDiffData %s from %s to %s (FileDiff %s)',
                           self.diff_hash.pk,
                           self.diff_hash.insert_count,
                           insert_count,
                           self.pk)

        if (delete_count is not None and
            raw_delete_count is not None and
            self.diff_hash.delete_count is not None and
            self.diff_hash.delete_count != delete_count):
            # Allow overriding, but warn. This really shouldn't be called.
            logger.warning('Attempting to override delete count on '
                           'RawFileDiffData %s from %s to %s (FileDiff %s)',
                           self.diff_hash.pk,
                           self.diff_hash.delete_count,
                           delete_count,
                           self.pk)

        if raw_insert_count is not None or raw_delete_count is not None:
            # New raw counts have been provided. These apply to the actual
            # diff file itself, and will be common across all diffs sharing
            # the diff_hash instance. Set it there.
            if raw_insert_count is not None:
                self.diff_hash.insert_count = raw_insert_count
                self.extra_data['raw_insert_count'] = raw_insert_count
                updated = True

            if raw_delete_count is not None:
                self.diff_hash.delete_count = raw_delete_count
                self.extra_data['raw_delete_count'] = raw_delete_count
                updated = True

            self.diff_hash.save()

        for key, cur_value in (('insert_count', insert_count),
                               ('delete_count', delete_count),
                               ('replace_count', replace_count),
                               ('equal_count', equal_count),
                               ('total_line_count', total_line_count)):
            if cur_value is not None and cur_value != self.extra_data.get(key):
                self.extra_data[key] = cur_value
                updated = True

        if updated and self.pk:
            self.save(update_fields=['extra_data'])

    def get_ancestors(self, minimal, filediffs=None, update=True):
        """Return the ancestors of this FileDiff.

        This will update the ancestors of this :py:class:`FileDiff` and all its
        ancestors if they are not already cached.

        Args:
            minimal (bool):
                Whether or not the minimal set of ancestors are returned.

                The minimal set of ancestors does not include ancestors where
                the file was deleted. In other words, if a file is deleted and
                then re-created, the deletion will not appear in the re-created
                file's ancestor list.

            filediffs (iterable of FileDiff, optional):
                An optional list of FileDiffs to check for ancestors so that a
                query does not have to be performed.

            update (bool, optional):
                Whether or not to cache the results in the database.

                If ``True`` and the results have not already been cached, this
                FileDiff and its ancestors will all be updated.

        Returns:
            list of FileDiff:
            The ancestor :py:class:`FileDiffs <FileDiff>`, in application
            order.
        """
        if self.commit_id is None:
            return []

        if (self.extra_data is None or
            self._ANCESTORS_KEY not in self.extra_data):
            if filediffs is None:
                filediffs = list(FileDiff.objects.filter(
                    diffset_id=self.diffset_id))

            compliment_ids, minimal_ids = self._compute_ancestors(filediffs,
                                                                  update)
        else:
            compliment_ids, minimal_ids = self.extra_data[self._ANCESTORS_KEY]

            if filediffs is None:
                filediffs = FileDiff.objects.filter(
                    pk__in=compliment_ids + minimal_ids)

        by_id = {
            filediff.pk: filediff
            for filediff in filediffs
        }

        if minimal:
            ids = minimal_ids
        else:
            ids = chain(compliment_ids, minimal_ids)

        return [by_id[pk] for pk in ids]

    def get_repository(self) -> Repository:
        """Return the repository this diff applies to.

        Version Added:
            4.0

        Returns:
            reviewboard.scmtools.models.Repository:
            The repository used for any operations for this FileDiff.
        """
        return self.diffset.repository

    def get_base_filediff(self, base_commit, ancestors=None):
        """Return the base FileDiff within a commit range.

        This looks through the FileDiff's ancestors, looking for the most
        recent one that equals or precedes ``base_commit``.

        Args:
            base_commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The newest commit that the resulting base FileDiff can be
                associated with.

            ancestors (list of FileDiff, optional):
                A pre-fetched list of FileDiff ancestors. If not provided, a
                list will be fetched.

        Returns:
            FileDiff:
            The base FileDiff, if one could be found. ``None`` will be
            returned if no base could be found.
        """
        if base_commit and self.commit_id:
            if ancestors is None:
                ancestors = self.get_ancestors(minimal=False) or []

            for ancestor in reversed(ancestors):
                if ancestor.commit_id <= base_commit.pk:
                    return ancestor

        return None

    def _compute_ancestors(self, filediffs, update):
        """Compute the ancestors of this FileDiff.

        Args:
            filediffs (iterable of FileDiff):
                The list of FileDiffs to check for ancestors.

            update (bool):
                Whether or not to cache the results.

        Returns:
            list of FileDiff:
            The ancestor FileDiffs in application order.
        """
        by_dest_file = {}
        by_id = {}

        for filediff in filediffs:
            by_detail = by_dest_file.setdefault(filediff.dest_file, {})
            by_commit = by_detail.setdefault(filediff.dest_detail, {})
            by_commit[filediff.commit_id] = filediff

            by_id[filediff.pk] = filediff

        current = self
        ancestor_ids = []
        should_update = {self.pk}

        while True:
            # If the ancestors have already been computed for the direct
            # ancestor, we can stop there and re-use those results.
            try:
                rest = current.extra_data[self._ANCESTORS_KEY]
            except (KeyError, TypeError):
                should_update.add(current.pk)
            else:
                # We reverse below, so we have to add these in reverse order.
                ancestor_ids.extend(chain(reversed(rest[1]),
                                          reversed(rest[0])))
                break

            if current.is_new:
                # If the FileDiff is new there may have been a previous
                # FileDiff with the same name that was deleted.
                try:
                    by_detail = by_dest_file[current.source_file]
                    prev_set = (
                        filediff
                        for by_commit in by_detail.values()
                        for filediff in by_commit.values()
                        if filediff.deleted
                    )
                except KeyError:
                    break
            else:
                # Otherwise we need to walk the graph to find the previous
                # FileDiff.
                try:
                    by_detail = by_dest_file[current.source_file]
                    prev_set = by_detail[current.source_revision].values()
                except KeyError:
                    # There is no previous FileDiff created by the commit
                    # series.
                    break

            # The only information we know is the previous revision and name,
            # of which there might be multiple matches. We need to find the
            # most recent FileDiff that matches that criteria that belongs to a
            # commit that comes before the current FileDiff's commit in
            # application order.
            try:
                prev = max(
                    (
                        filediff
                        for filediff in prev_set
                        if filediff.commit_id < current.commit_id
                    ),
                    key=lambda filediff: filediff.commit_id)
            except ValueError:
                # max() raises ValueError if it is given an empty iterable.
                # This means there is no previous FileDiff created by the
                # commit series.
                break

            ancestor_ids.append(prev.pk)
            current = prev

        # We computed the list of ancestors in reverse order (i.e., most recent
        # to least recent) above.
        ancestor_ids.reverse()

        compliment_ids, minimal_ids = self._split_ancestors(by_id,
                                                            ancestor_ids)

        if update:
            for i, pk in enumerate(ancestor_ids):
                if pk in should_update:
                    filediff = by_id[pk]

                    if filediff.extra_data is None:
                        filediff.extra_data = {}

                    # We need to split the history at the point of the last
                    # deletion. That way we will have the minimal set of
                    # ancestors, which we can use to compute the diff for this
                    # FileDiff, and the maximal set of ancestors, which we can
                    # use to compute cumulative diffs.
                    filediff.extra_data[self._ANCESTORS_KEY] = \
                        list(self._split_ancestors(by_id, ancestor_ids[:i]))
                    filediff.save(update_fields=('extra_data',))

            self.extra_data[self._ANCESTORS_KEY] = [compliment_ids,
                                                    minimal_ids]
            self.save(update_fields=('extra_data',))

        return compliment_ids, minimal_ids

    def _split_ancestors(self, by_id, ancestor_ids):
        """Split the ancestor IDs into pre-delete and post-delete sets.

        Args:
            by_id (dict):
                A mapping of primary keys to :py:class:`FileDiffs
                <reviewboard.diffviewer.models.filediff.FileDiff>`.

            ancestor_ids (list of int):
                The ancestors to split.

        Returns:
            tuple:
            A 2-tuple of:

            * The compliment of the minimal ancestors (py:class:`list` of
              :py:class:`int`).
            * The list of minimal ancestors (py:class:`list` of
              :py:class:`int`).
        """
        i = 0

        # We traverse the list backwards to find the last deleted FileDiff.
        # Everything after that is in the set of minimal ancestors.
        for i in range(len(ancestor_ids) - 1, -1, -1):
            filediff = by_id[ancestor_ids[i]]

            if filediff.deleted:
                i += 1
                break

        return ancestor_ids[:i], ancestor_ids[i:]

    def _needs_diff_migration(self):
        return self.diff_hash_id is None

    def _needs_parent_diff_migration(self):
        return (self.parent_diff_hash_id is None and
                (self.parent_diff64 or self.legacy_parent_diff_hash_id))

    def _migrate_diff_data(self, recalculate_counts=True):
        """Migrates diff data associated with a FileDiff to RawFileDiffData.

        If the diff data is stored directly on the FileDiff, it will be
        removed and stored on a RawFileDiffData instead.

        If the diff data is stored on an associated LegacyFileDiffData,
        that will be converted into a RawFileDiffData. The LegacyFileDiffData
        will then be removed, if nothing else is using it.

        Args:
            recalculate_line_counts (bool, optional):
                Whether or not line counts should be recalculated during the
                migration.
        """
        needs_save = False
        diff_hash_is_new = False
        parent_diff_hash_is_new = False
        fix_refs = False
        legacy_pks = []
        needs_diff_migration = self._needs_diff_migration()
        needs_parent_diff_migration = self._needs_parent_diff_migration()

        if needs_diff_migration:
            recalculate_diff_counts = recalculate_counts
            needs_save = True

            if self.legacy_diff_hash_id:
                logger.debug('Migrating LegacyFileDiffData %s to '
                             'RawFileDiffData for diff in FileDiff %s',
                             self.legacy_diff_hash_id, self.pk)

                try:
                    legacy_data = self.legacy_diff_hash.binary
                except LegacyFileDiffData.DoesNotExist:
                    # Another process migrated this before we could.
                    # We'll need to fix the references.
                    fix_refs = True
                    recalculate_diff_counts = False
                else:
                    diff_hash_is_new = self._set_diff(legacy_data)
                    legacy_pks.append(self.legacy_diff_hash_id)
                    self.legacy_diff_hash = None
            else:
                logger.debug('Migrating FileDiff %s diff data to '
                             'RawFileDiffData',
                             self.pk)

                diff_hash_is_new = self._set_diff(self.diff64)

            if recalculate_diff_counts:
                self._recalculate_line_counts(self.diff_hash)

        if needs_parent_diff_migration:
            recalculate_parent_diff_counts = recalculate_counts
            needs_save = True

            if self.legacy_parent_diff_hash_id:
                logger.debug('Migrating LegacyFileDiffData %s to '
                             'RawFileDiffData for parent diff in FileDiff %s',
                             self.legacy_parent_diff_hash_id, self.pk)

                try:
                    legacy_parent_data = self.legacy_parent_diff_hash.binary
                except LegacyFileDiffData.DoesNotExist:
                    # Another process migrated this before we could.
                    # We'll need to fix the references.
                    fix_refs = True
                    recalculate_parent_diff_counts = False
                else:
                    parent_diff_hash_is_new = \
                        self._set_parent_diff(legacy_parent_data)
                    legacy_pks.append(self.legacy_parent_diff_hash_id)
                    self.legacy_parent_diff_hash = None
            else:
                logger.debug('Migrating FileDiff %s parent diff data to '
                             'RawFileDiffData',
                             self.pk)

                parent_diff_hash_is_new = \
                    self._set_parent_diff(self.parent_diff64)

            if recalculate_parent_diff_counts:
                self._recalculate_line_counts(self.parent_diff_hash)

        if fix_refs:
            # Another server/process/thread got to this before we could.
            # We need to pull the latest refs and make sure they're set here.
            diff_hash, parent_diff_hash = (
                FileDiff.objects.filter(pk=self.pk)
                .values_list('diff_hash_id', 'parent_diff_hash_id')[0]
            )

            if needs_diff_migration:
                if diff_hash:
                    self.diff_hash_id = diff_hash
                    self.legacy_diff_hash = None
                    self.diff64 = ''
                else:
                    logger.error('Unable to migrate diff for FileDiff %s: '
                                 'LegacyFileDiffData "%s" is missing, and '
                                 'database entry does not have a new '
                                 'diff_hash! Data may be missing.',
                                 self.pk, self.legacy_diff_hash_id)

            if needs_parent_diff_migration:
                if parent_diff_hash:
                    self.parent_diff_hash_id = parent_diff_hash
                    self.legacy_parent_diff_hash = None
                    self.parent_diff64 = ''
                else:
                    logger.error('Unable to migrate parent diff for '
                                 'FileDiff %s: LegacyFileDiffData "%s" is '
                                 'missing, and database entry does not have '
                                 'a new parent_diff_hash! Data may be '
                                 'missing.',
                                 self.pk, self.legacy_parent_diff_hash_id)

        if needs_save:
            if self.pk:
                self.save(update_fields=[
                    'diff64', 'parent_diff64', 'diff_hash', 'parent_diff_hash',
                    'legacy_diff_hash', 'legacy_parent_diff_hash',
                ])
            else:
                self.save()

        if legacy_pks:
            # Delete any LegacyFileDiffData objects no longer associated
            # with any FileDiffs.
            LegacyFileDiffData.objects \
                .filter(pk__in=legacy_pks) \
                .exclude(Q(filediffs__pk__gt=0) |
                         Q(parent_filediffs__pk__gt=0)) \
                .delete()

        return diff_hash_is_new, parent_diff_hash_is_new

    def _recalculate_line_counts(self, diff_hash):
        """Recalculate line counts for the raw data.

        Args:
            diff_hash (reviewboard.diffviewer.models.raw_file_diff_data.
                       RawFileDiffData):
                The raw data to recalculate line counts for.
        """
        diff_hash.recalculate_line_counts(self.get_repository().get_scmtool())

    def get_review_request(self) -> Optional[ReviewRequest]:
        """Return the ReviewRequest that this filediff is attached to.

        Version Added:
            7.0

        Returns:
            reviewboard.reviews.models.review_request.ReviewRequest:
            The review request.
        """
        if not hasattr(self, '_review_request'):
            diffset = self.diffset

            if diffset.history:
                self._review_request = diffset.history.review_request.get()
            else:
                try:
                    draft = diffset.review_request_draft.get()
                    self._review_request = draft.review_request
                except ObjectDoesNotExist:
                    # This should not be hit in real-world usage, since
                    # FileDiffs really ought to be connected to a review
                    # request in some form or another. In any case, we don't
                    # cache the result for this because maybe we'll attach the
                    # FileDiff to a ReviewRequest later in this same request.
                    return None

        return self._review_request

    def __str__(self):
        """Return a human-readable representation of the model.

        Returns:
            unicode:
            A human-readable representation of the model.
        """
        return "%s (%s) -> %s (%s)" % (self.source_file, self.source_revision,
                                       self.dest_file, self.dest_detail)

    class Meta:
        app_label = 'diffviewer'
        db_table = 'diffviewer_filediff'
        verbose_name = _('File Diff')
        verbose_name_plural = _('File Diffs')



"""Views for handling authentication and user accounts."""

from __future__ import annotations

import logging
from typing import Optional
from urllib.parse import quote, urlparse

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.contrib.auth.views import (
    LoginView as DjangoLoginView,
    LogoutView,
    logout_then_login as auth_logout_then_login)
from django.forms.forms import ErrorDict
from django.http import (Http404,
                         HttpRequest,
                         HttpResponse,
                         HttpResponseRedirect)
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.html import escape
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_protect
from django.views.generic.base import TemplateView
from djblets.auth.views import register
from djblets.configforms.views import ConfigPagesView
from djblets.features.decorators import feature_required
from djblets.forms.fieldsets import filter_fieldsets
from djblets.registries.errors import ItemLookupError
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import augment_method_from
from djblets.views.generic.etag import ETagViewMixin
from djblets.webapi.errors import WebAPITokenGenerationError

from reviewboard.accounts.backends import get_enabled_auth_backends
from reviewboard.accounts.forms.registration import RegistrationForm
from reviewboard.accounts.mixins import (CheckLoginRequiredViewMixin,
                                         LoginRequiredViewMixin)
from reviewboard.accounts.pages import AccountPage, OAuth2Page, PrivacyPage
from reviewboard.accounts.privacy import is_consent_missing
from reviewboard.accounts.sso.backends import sso_backends
from reviewboard.admin.decorators import check_read_only
from reviewboard.avatars import avatar_services
from reviewboard.notifications.email.decorators import preview_email
from reviewboard.notifications.email.message import \
    prepare_password_changed_mail
from reviewboard.oauth.features import oauth2_service_feature
from reviewboard.oauth.forms import (UserApplicationChangeForm,
                                     UserApplicationCreationForm)
from reviewboard.oauth.models import Application
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.webapi.models import WebAPIToken


logger = logging.getLogger(__name__)


class LoginView(DjangoLoginView):
    """A view for rendering the login page.

    This view may be called when clients are trying to authenticate to
    Review Board through a web-based login flow. In that case, callers must
    include a ``client-name`` query parameter containing the client name,
    and a ``client-url`` parameter containing the URL of where to send
    authentication data upon a successful login. Client callers may include a
    ``next`` parameter containing a URL for redirection, making sure to
    encode any query parameters in that URL.

    Version Changed:
        5.0.5:
        Added the ``client-name`` and ``client-url`` query parameters for
        authenticating clients.

    Version Added:
        5.0

    Version Changed:
        5.0.5:
        Added the ``client-name`` and ``client-url`` query parameters for
        authenticating clients.
    """

    template_name = 'accounts/login.html'

    ######################
    # Instance variables #
    ######################

    #: Whether the request is for authenticating a client.
    #:
    #: Version Added:
    #:     5.0.5
    #:
    #: Type:
    #:     bool
    client_auth_flow: bool

    #: The name of the client who is authenticating.
    #:
    #: Version Added:
    #:     5.0.5
    #:
    #: Type:
    #:     str
    client_name: Optional[str]

    #: The URL to the client login page.
    #:
    #: Version Added:
    #:     5.0.5
    #:
    #: Type:
    #:     str
    client_login_url: str

    #: The URL to the client login confirmation page.
    #:
    #: Version Added:
    #:     5.0.5
    #:
    #: Type:
    #:     str
    client_login_confirm_url: str

    #: The URL of where to send authentication data for the client.
    #:
    #: Version Added:
    #:     5.0.5
    #:
    #: Type:
    #:     str
    client_url: Optional[str]

    def dispatch(self, request, *args, **kwargs):
        """Dispatch the view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple):
                Positional arguments to pass through to the parent class.

            **kwargs (dict):
                Keyword arguments to pass through to the parent class.

        Returns:
            django.http.HttpResponse:
            The response to send to the client.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        self.client_name = None
        self.client_url = None

        if siteconfig.get('client_web_login'):
            self.client_name = self.request.GET.get(
                'client-name',
                self.request.POST.get('client-name', ''))
            self.client_url = self.request.GET.get(
                'client-url',
                self.request.POST.get('client-url', ''))
            client_url_port = urlparse(self.client_url).port
            self.success_url_allowed_hosts = \
                _get_client_allowed_hosts(client_url_port)

        client_name = self.client_name
        client_url = self.client_url
        client_auth_flow = bool(client_name and client_url)
        self.client_auth_flow = client_auth_flow
        client_redirect_param_str = ''
        redirect_field_name = self.redirect_field_name
        redirect_to = self.get_redirect_url()

        if redirect_to and client_auth_flow:
            client_redirect_param_str = (
                '&%s=%s' % (redirect_field_name, quote(redirect_to)))

        client_login_url = (
            '%s?client-name=%s&client-url=%s%s'
            % (local_site_reverse('client-login'),
               client_name,
               client_url,
               client_redirect_param_str))
        client_login_confirm_url = (
            '%s?client-name=%s&client-url=%s%s'
            % (local_site_reverse('client-login-confirm'),
               client_name,
               client_url,
               client_redirect_param_str))
        self.client_login_url = client_login_url
        self.client_login_confirm_url = client_login_confirm_url

        if (request.method == 'GET' and client_auth_flow and
            request.user.is_authenticated):
            # The request is for client web-based login, with the user already
            # logged in.
            return HttpResponseRedirect(client_login_confirm_url)

        sso_auto_login_backend = siteconfig.get('sso_auto_login_backend', None)

        if sso_auto_login_backend:
            try:
                backend = sso_backends.get('backend_id', sso_auto_login_backend)
                login_url = backend.login_url

                if client_auth_flow:
                    redirect_to = client_login_confirm_url
                else:
                    redirect_to = self.get_success_url()

                if url_has_allowed_host_and_scheme(
                    url=redirect_to, allowed_hosts=request.get_host()):
                    login_url = '%s?%s=%s' % (login_url,
                                              redirect_field_name,
                                              quote(redirect_to))

                return HttpResponseRedirect(login_url)
            except ItemLookupError:
                logger.error('Unable to find sso_auto_login_backend "%s".',
                             sso_auto_login_backend)

        if request.method == 'GET' and request.user.is_authenticated:
            # If this is a normal GET request and not part of any flow, and
            # the user is already logged in, just redirect them. Otherwise
            # if they're on the Log In page, click Log In, and get a redirect
            # URL back to the Log In page, a successful login will look like
            # failure, as they'll just get taken back here. This has come up
            # with a user, and is indeed confusing.
            #
            # Note that Django lets us set `redirect_authenticated_user = True`
            # to get this behavior, but it will handle a redirect URL of the
            # login page ungracefully (raising an exception that becomes a
            # HTTP 500). We're handling it ourselves instead.
            return HttpResponseRedirect(self.get_success_url())

        return super().dispatch(request, *args, **kwargs)

    def get_success_url(self) -> str:
        """Return the direct URL for a successful login.

        This overrides the default behavior to fall back to the root for the
        global site or the accessed Local Site.

        Returns:
            str:
            The resulting redirect URL.
        """
        return (self.get_redirect_url() or
                local_site_reverse('root', request=self.request))

    def get_redirect_url(self) -> str:
        """Return the user-provided redirect URL for this request.

        This will determine the requested redirect URL, if any, and then
        normalize it to avoid any login -> logout redirects or login -> login
        redirects.

        Returns:
            str:
            The resulting redirect URL.
        """
        redirect_to = super().get_redirect_url()
        request = self.request

        if (redirect_to == request.path or
            redirect_to == local_site_reverse('logout', request=request)):
            # We're in a redirect loop or a login -> logout loop. Just ignore
            # the user-supplied redirect URL.
            redirect_to = ''

        return redirect_to

    def get_context_data(self, **kwargs):
        """Return extra data for rendering the template.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the parent class.

        Returns:
            dict:
            Context to use when rendering the template.
        """
        context = super().get_context_data(**kwargs)

        context['enabled_sso_backends'] = [
            sso_backend
            for sso_backend in sso_backends
            if sso_backend.is_enabled()
        ]

        if self.client_auth_flow:
            context['client_name'] = self.client_name
            context['client_url'] = self.client_url

            # While in the client web-based login flow, redirect to
            # the client login page upon successful login.
            context[self.redirect_field_name] = self.client_login_url

        return context


def logout(request, *args, **kwargs):
    """Log out the user."""
    siteconfig = SiteConfiguration.objects.get_current()
    sso_auto_login_backend = siteconfig.get('sso_auto_login_backend', None)

    # If we're configured to automatically log in via SSO, we can't use
    # logout_then_login, because it will log out and then immediately log in
    # again.
    if sso_auto_login_backend:
        try:
            sso_backends.get('backend_id', sso_auto_login_backend)
            return LogoutView.as_view()(request, *args, **kwargs)
        except ItemLookupError:
            logger.error('Unable to find sso_auto_login_backend "%s".',
                         sso_auto_login_backend)

    return auth_logout_then_login(request, *args, **kwargs)


class UserInfoboxView(CheckLoginRequiredViewMixin,
                      CheckLocalSiteAccessViewMixin,
                      ETagViewMixin,
                      TemplateView):
    """Displays information on a user, for use in user pop-up infoboxes.

    This is meant to be embedded in other pages, rather than being
    a standalone page.
    """

    template_name = 'accounts/user_infobox.html'

    def __init__(self, **kwargs):
        """Initialize a view for the request.

        Args:
            **kwargs (dict):
                Keyword arguments passed to :py:meth:`as_view`.
        """
        super(UserInfoboxView, self).__init__(**kwargs)

        self._lookup_user = None
        self._show_profile = None
        self._timezone = None

    def get_etag_data(self, request, username, *args, **kwargs):
        """Return an ETag for the view.

        This will look up some state needed for the request and generate a
        suitable ETag.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            username (unicode):
                The username of the user being looked up.

            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (tuple):
                Keyword arguments to pass to the handler.

                These will be arguments provided by the URL pattern.

        Returns:
            unicode:
            The ETag for the page.
        """
        from reviewboard.extensions.hooks import UserInfoboxHook

        user = get_object_or_404(User, username=username)
        self._lookup_user = user

        profile = user.get_profile()
        self._show_profile = user.is_profile_visible(request.user)
        self._timezone = profile.timezone

        etag_data = [
            user.first_name,
            user.last_name,
            user.email,
            str(user.last_login),
            str(settings.TEMPLATE_SERIAL),
            str(self._show_profile),
            self._timezone,
        ]

        if avatar_services.avatars_enabled:
            avatar_service = avatar_services.for_user(user)

            if avatar_service:
                etag_data.extend(avatar_service.get_etag_data(user))

        local_site = self.local_site

        for hook in UserInfoboxHook.hooks:
            try:
                etag_data.append(hook.get_etag_data(
                    user=user,
                    request=request,
                    local_site=local_site))
            except Exception as e:
                logger.exception('Error when running UserInfoboxHook.'
                                 'get_etag_data method in extension "%s": %s',
                                 hook.extension.id, e,
                                 extra={'request': request})

        return ':'.join(etag_data)

    def get_context_data(self, **kwargs):
        """Return data for the template.

        This will return information on the user, along with information from
        any extension hooks used for the page.

        Args:
            **kwargs (tuple):
                Additional keyword arguments from the URL pattern.

        Returns:
            dict:
            Context data for the template.
        """
        from reviewboard.extensions.hooks import UserInfoboxHook

        # These are accessed several times, so bring them in to reduce
        # attribute lookups.
        user = self._lookup_user
        username = user.username
        local_site = self.local_site

        extra_content = []

        for hook in UserInfoboxHook.hooks:
            try:
                extra_content.append(hook.render(
                    user=user,
                    request=self.request,
                    local_site=local_site))
            except Exception as e:
                logger.exception('Error when running UserInfoboxHook.'
                                 'render method in extension "%s": %s',
                                 hook.extension.id, e,
                                 extra={'request': self.request})

        review_requests_url = local_site_reverse('user', local_site=local_site,
                                                 args=[username])
        reviews_url = local_site_reverse('user-grid', local_site=local_site,
                                         args=[username, 'reviews'])

        has_avatar = (
            avatar_services.avatars_enabled and
            avatar_services.for_user(user) is not None
        )

        return {
            'extra_content': mark_safe(''.join(extra_content)),
            'full_name': user.get_full_name(),
            'has_avatar': has_avatar,
            'infobox_user': user,
            'review_requests_url': review_requests_url,
            'reviews_url': reviews_url,
            'show_profile': self._show_profile,
            'timezone': self._timezone,
        }


@csrf_protect
def account_register(request, next_url='dashboard'):
    """Display the appropriate registration page.

    If registration is enabled and the selected authentication backend supports
    creation of users, this will return the appropriate registration page. If
    registration is not supported, this will redirect to the login view.
    """
    siteconfig = SiteConfiguration.objects.get_current()
    auth_backends = get_enabled_auth_backends()

    if (auth_backends[0].supports_registration and
        siteconfig.get('auth_enable_registration') and
        not siteconfig.get('site_read_only')):
        response = register(request, next_page=reverse(next_url),
                            form_class=RegistrationForm)

        return response

    return HttpResponseRedirect(reverse("login"))


class MyAccountView(ConfigPagesView):
    """Displays the My Account page containing user preferences.

    The page will be built based on registered pages and forms. This makes
    it easy to plug in new bits of UI for the page, which is handy for
    extensions that want to offer customization for users.
    """

    title = _('My Account')

    css_bundle_names = [
        'account-page',
    ]

    js_bundle_names = [
        '3rdparty-jsonlint',
        'config-forms',
        'account-page',
    ]

    @method_decorator(login_required)
    @method_decorator(check_read_only)
    @augment_method_from(ConfigPagesView)
    def dispatch(self, *args, **kwargs):
        """Handle the view.

        This just falls back to the djblets ConfigPagesView.dispatch
        implementation.
        """
        pass

    @property
    def nav_title(self):
        """Get the title for the navigation section."""
        return self.request.user.username

    @property
    def page_classes(self):
        """The list of page classes for this view.

        If the user is missing any consent requirements or has not accepted
        the privacy policy/terms of service, only the privacy page will be
        shown.
        """
        if self.is_user_missing_consent:
            return [AccountPage.registry.get('page_id', PrivacyPage.page_id)]

        return list(AccountPage.registry)

    @cached_property
    def ordered_user_local_sites(self):
        """Get the user's local sites, ordered by name."""
        return self.request.user.local_site.order_by('name')

    @property
    def render_sidebar(self):
        """Whether or not to render the sidebar.

        If the user is missing any consent requirements or has not accepted
        the privacy policy/terms of service, the sidebar will not render.
        This is to prevent the user from navigating away from the privacy page
        before making decisions.
        """
        return not self.is_user_missing_consent

    @cached_property
    def is_user_missing_consent(self):
        """Whether or not the user is missing consent."""
        return is_consent_missing(self.request.user)


@login_required
@preview_email(prepare_password_changed_mail)
def preview_password_changed_email(request):
    return {
        'user': request.user,
    }


@login_required
@feature_required(oauth2_service_feature)
def edit_oauth_app(request, app_id=None):
    """Create or edit an OAuth2 application.

    Args:
        request (django.http.HttpRequest):
            The current HTTP request.

        app_id (int, optional):
            The ID of the application to edit.

            If this argument is ``None`` a new application will be edited.

    Returns:
        django.http.HttpResponse:
        The rendered view.
    """
    # If we import this at global scope, it will cause issues with admin sites
    # being automatically registered.
    from reviewboard.oauth.admin import ApplicationAdmin

    if app_id:
        app = get_object_or_404(
            Application,
            pk=app_id,
            user=request.user,
        )
        form_cls = UserApplicationChangeForm
        fieldsets = ApplicationAdmin.fieldsets
    else:
        app = None
        form_cls = UserApplicationCreationForm
        fieldsets = ApplicationAdmin.add_fieldsets

    if request.method == 'POST':
        form_data = request.POST.copy()

        form = form_cls(user=request.user, data=form_data, initial=None,
                        instance=app)

        if form.is_valid():
            app = form.save()

            if app_id is not None:
                next_url = OAuth2Page.get_absolute_url()
            else:
                next_url = reverse('edit-oauth-app', args=(app.pk,))

            return HttpResponseRedirect(next_url)
    else:
        form = form_cls(user=request.user, data=None, initial=None,
                        instance=app)

        # Show a warning at the top of the form when the form is disabled for
        # security.
        #
        # We don't need to worry about full_clean not being called (which would
        # be if we went through form.errors) because this form will never be
        # saved.
        if app and app.is_disabled_for_security:
            form._errors = ErrorDict({
                '__all__': form.error_class(
                    [form.DISABLED_FOR_SECURITY_ERROR],
                ),
            })

    return render(
        request=request,
        template_name='accounts/edit_oauth_app.html',
        context={
            'app': app,
            'form': form,
            'fieldsets': filter_fieldsets(form=form_cls,
                                          fieldsets=fieldsets),
            'oauth2_page_url': OAuth2Page.get_absolute_url(),
            'request': request,
        })


class BaseClientLoginView(LoginRequiredViewMixin,
                          TemplateView):
    """Base view for views dealing with the client web-based login flow.

    Callers must include a ``client-name`` query parameter containing the
    client name, and a ``client-url`` parameter containing the URL of where
    to send authentication data upon a successful login. Callers may include
    a ``next`` parameter containing a URL for redirection, making sure to
    encode any query parameters in that URL.

    Version Added:
        5.0.5
    """

    #: The name of the redirect field.
    #:
    #: Type:
    #:     str
    redirect_field_name: str = LoginView.redirect_field_name

    ######################
    # Instance variables #
    ######################

    #: Whether the client is safe to redirect and POST to.
    #:
    #: Type:
    #:     bool
    client_allowed: bool

    #: The hosts that are allowed to authenticate clients to Review Board.
    #:
    #: This will only allow the local host server at the given client
    #: port and the Review Board server.
    #:
    #: Type:
    #:     set
    client_allowed_hosts: set

    #: The name of the client who is authenticating.
    #:
    #: Type:
    #:     str
    client_name: str

    #: The URL of where to send authentication data for the client.
    #:
    #: Type:
    #:     str
    client_url: str

    #: The URL of where to redirect to upon a successful login.
    #:
    #: This is URL encoded.
    #:
    #: Type:
    #:     str
    redirect_to: str

    #: The name of the template to render.
    #:
    #: This must be set by the subclass.
    #:
    #: Type:
    #:     str
    template_name: str

    def dispatch(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Dispatch the view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple):
                Positional arguments to pass through to the parent class.

            **kwargs (dict):
                Keyword arguments to pass through to the parent class.

        Returns:
            django.http.HttpResponse:
            The response to send to the client.
        """
        self.siteconfig = SiteConfiguration.objects.get_current()
        request_GET = self.request.GET

        if self.siteconfig.get('client_web_login'):
            self.client_name = request_GET.get('client-name', '')
            self.client_url = request_GET.get('client-url', '')

            client_url_port = urlparse(self.client_url).port
            client_allowed_hosts = _get_client_allowed_hosts(client_url_port)
            client_allowed_hosts.add(request.get_host())
            self.client_allowed_hosts = client_allowed_hosts

            self.client_allowed = self._url_is_safe(self.client_url)
            self.redirect_to = self._get_redirect_url()

            return super().dispatch(request, *args, **kwargs)

        raise Http404

    def get_context_data(self, **kwargs) -> dict:
        """Return extra data for rendering the template.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the parent class.

        Returns:
            dict:
            Context to use when rendering the template.
        """
        context = super().get_context_data(**kwargs)

        context['client_allowed'] = self.client_allowed
        context['client_name'] = self.client_name
        context['client_url'] = self.client_url
        context['username'] = self.request.user.username

        return context

    def _url_is_safe(
        self,
        url: str,
    ) -> bool:
        """Return whether the given URL is safe to redirect and/or POST to.

        Args:
            url (str):
                The URL.

        Returns:
            bool:
            Whether the url is safe to redirect and/or POST to.
        """
        return url_has_allowed_host_and_scheme(
            url=url,
            allowed_hosts=self.client_allowed_hosts)

    def _get_redirect_url(self) -> str:
        """Return the redirect URL.

        This encodes the URL.

        Returns:
            str:
            The redirect URL or an empty string if the redirect URL is
            not safe.
        """
        redirect_to = self.request.POST.get(
            self.redirect_field_name,
            self.request.GET.get(self.redirect_field_name, ''))
        assert isinstance(redirect_to, str)

        if self._url_is_safe(redirect_to):
            return quote(redirect_to)

        return ''


class ClientLoginView(BaseClientLoginView):
    """View for rendering the client login page.

    The client login page handles authenticating a client to Review Board
    by POSTing authentication data to the client.

    Callers must include a ``client-name`` query parameter containing the
    client name, and a ``client-url`` parameter containing the URL of where
    to send authentication data upon a successful login. Callers may include
    a ``next`` parameter containing a URL for redirection, making sure to
    encode any query parameters in that URL.

    Version Added:
        5.0.5
    """

    template_name = 'accounts/client_login.html'

    def get_context_data(self, **kwargs) -> dict:
        """Return extra data for rendering the template.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the parent class.

        Returns:
            dict:
            Context to use when rendering the template.
        """
        context = super().get_context_data(**kwargs)

        context['js_view_data'] = self.get_js_view_data()
        error = context['js_view_data'].pop('error', '')

        if error:
            context['error'] = error

        return context

    def get_js_view_data(self) -> dict:
        """Return the data for the ClientLoginView JavaScript view.

        Returns:
            dict:
            Data to be passed to the JavaScript view.
        """
        client_allowed = self.client_allowed
        client_name = self.client_name
        client_url = self.client_url
        payload = {}
        error = ''

        if client_allowed:
            try:
                api_token = WebAPIToken.objects.get_or_create_client_token(
                    client_name=client_name,
                    user=self.request.user)[0]

                payload = {
                    'api_token': api_token.token,
                }
            except WebAPITokenGenerationError:
                error = _(
                    'Failed to generate a unique API token for '
                    'authentication. Please reload the page to try again.')
        else:
            logger.warning('Blocking an attempt to send authentication info '
                           'to unsafe URL %s', client_url)
        return {
            'clientName': escape(client_name),
            'clientURL': quote(client_url),
            'error': error,
            'payload': payload,
            'redirectTo': self.redirect_to,
            'username': self.request.user.username,
        }


class ClientLoginConfirmationView(BaseClientLoginView):
    """View for rendering the client login confirmation page.

    This page asks the user if they want to authenticate the client as
    the current user who is logged in to Review Board. If yes, they will be
    redirected to the client login page. If not, they will be logged
    out and redirected to the login page.

    Callers must include a ``client-name`` query parameter containing the
    client name, and a ``client-url`` parameter containing the URL of where
    to send authentication data upon a successful login. Callers may include
    a ``next`` parameter containing a URL for redirection, making sure to
    encode any query parameters in that URL.

    Version Added:
        5.0.5
    """

    template_name = 'accounts/client_login_confirm.html'

    def get_context_data(self, **kwargs) -> dict:
        """Return extra data for rendering the template.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the parent class.

        Returns:
            dict:
            Context to use when rendering the template.
        """
        context = super().get_context_data(**kwargs)

        client_name = self.client_name
        client_url = self.client_url
        redirect_field_name = self.redirect_field_name
        redirect_to = self.redirect_to
        client_redirect_param_str = ''

        if redirect_to:
            client_redirect_param_str = (
                '&%s=%s' % (redirect_field_name, redirect_to))

        context['client_login_url'] = (
            '%s?client-name=%s&client-url=%s%s'
            % (local_site_reverse('client-login'),
               client_name,
               client_url,
               client_redirect_param_str))

        # The client redirect part of the URL is encoded twice
        # in order to preserve all of its query parameters.
        logout_redirect = quote(
            '%s?client-name=%s&client-url=%s%s'
            % (local_site_reverse('login'),
               client_name,
               client_url,
               client_redirect_param_str))
        context['logout_url'] = (
            '%s?%s=%s'
            % (local_site_reverse('logout'),
               redirect_field_name,
               logout_redirect))

        return context


def _get_client_allowed_hosts(
    port: Optional[int],
) -> set:
    """Return the set of hosts that are allowed to authenticate clients.

    This will return a set of the local host names at the given port,
    or with no port specified if one is not given.

    Version Added:
        5.0.5

    Args:
        port (int):
            The specific port to allow for the hosts. If this is
            ``None`` then no port will be specified.

    Returns:
        set:
        The set of allowed hosts.
    """
    if port:
        suffix = f':{port}'
    else:
        suffix = ''

    return {
        f'127.0.0.1{suffix}',
        f'localhost{suffix}'
    }
